"""
Module contains validator classes and predicate functions useful for validation.
Example of use (how things play together):
    >>> option_dict = {"some_option": "A"}
    >>> validators = [
    ...     IsRequiredAll(["name"]),
    ...     ValueIn("some_option", ["B", "C"]),
    ... ]
    >>> report_list = ValidatorAll(validators).validate(option_dict)
    >>> for report in report_list:
    ...     print(report)
    ...
    ...
    ERROR REQUIRED_OPTION_IS_MISSING: {
        'option_type': 'option',
        'option_names': ['name']
    }
    ERROR INVALID_OPTION_VALUE: {
        'option_name': 'some_option',
        'option_value': 'A',
        'allowed_values': ['B', 'C']
    }

ValuePair class and values_to_pairs and pairs_to_values helpers are usefull in
cases when we walidate a normalized value. If the normalized value is not
valid, we want to put the original value into resulting reports. This is to
prevent confusion which may happen if the normalized value different from the
entered one is reported as not valid.
"""
from collections import namedtuple
import ipaddress
import re

from pcs.lib import reports
from pcs.lib.corosync import constants as corosync_constants
from pcs.lib.errors import ReportItemSeverity
from pcs.lib.pacemaker.values import (
    timeout_to_seconds,
    validate_id,
)

_INTEGER_RE = re.compile(r"^[+-]?[0-9]+$")

### normalization

class ValuePair(namedtuple("ValuePair", "original normalized")):
    """
    Storage for the original value and its normalized form
    """
    @staticmethod
    def get(val):
        return val if isinstance(val, ValuePair) else ValuePair(val, val)

def values_to_pairs(option_dict, normalize):
    """
    Return a dict derived from option_dict where every value is instance of
    ValuePair.

    dict option_dict contains values that should be paired with the normalized
        form
    callable normalize should take key and value and return normalized form.
        Function option_value_normalization can be good base for create such
        callable.
    """
    option_dict_with_pairs = {}
    for key, value in option_dict.items():
        if not isinstance(value, ValuePair):
            value = ValuePair(
                original=value,
                normalized=normalize(key, value),
            )
        option_dict_with_pairs[key] = value
    return option_dict_with_pairs

def pairs_to_values(option_dict):
    """
    Take a dict which has OptionValuePairs as its values and return dict with
    normalized forms as its values. It is reverse function to
    values_to_pairs.

    dict option_dict contains OptionValuePairs as its values
    """
    raw_option_dict = {}
    for key, value in option_dict.items():
        if isinstance(value, ValuePair):
            value = value.normalized
        raw_option_dict[key] = value
    return raw_option_dict

def option_value_normalization(normalization_map):
    """
    Return function that takes key and value and return the normalized form.

    dict normalization_map has on each key function that takes value and return
        its normalized form.
    """
    def normalize(key, value):
        return(
            value if key not in normalization_map
            else normalization_map[key](value)
        )
    return normalize

### generic validators

class ValidatorInterface():
    """
    Base interface of all validators
    """
    def validate(self, option_dict):
        raise NotImplementedError()

class CompoundValidator(ValidatorInterface):
    """
    Base abstract class for compound validators
    """
    def __init__(self, validator_list):
        self._validator_list = validator_list

    def validate(self, option_dict):
        raise NotImplementedError()

class ValidatorAll(CompoundValidator):
    """
    Run all validators and return all their reports
    """
    def validate(self, option_dict):
        report_list = []
        for validator in self._validator_list:
            report_list.extend(validator.validate(option_dict))
        return report_list

class ValidatorFirstError(CompoundValidator):
    """
    Run validators in sequence, return reports once one reports an error
    """
    def validate(self, option_dict):
        report_list = []
        for validator in self._validator_list:
            new_report_list = validator.validate(option_dict)
            report_list.extend(new_report_list)
            error_reported = False
            for report_item in new_report_list:
                if report_item.severity == ReportItemSeverity.ERROR:
                    error_reported = True
                    break
            if error_reported:
                break
        return report_list

### keys validators

class KeyValidator(ValidatorInterface):
    def __init__(self, option_name_list, option_type=None):
        """
        iterable option_name_list -- names of the options to check
        string option_type -- describes a type of the options for reporting
        """
        self._option_name_list = option_name_list
        self._option_type = option_type

    def validate(self, option_dict):
        raise NotImplementedError()

class CorosyncOption(KeyValidator):
    """
    Report INVALID_USERDEFINED_OPTIONS when the option_dict contains names not
    suitable for corosync.conf
    """
    def __init__(self, option_type=None):
        super().__init__(None, option_type=option_type)

    def validate(self, option_dict):
        not_valid_options = [
            name
            for name in option_dict
            if corosync_constants.OPTION_NAME_RE.fullmatch(name) is None
        ]
        if not_valid_options:
            # We must be strict and do not allow to override this validation,
            # otherwise setting a cratfed option name could be misused for
            # setting arbitrary corosync.conf settings.
            return [
                reports.invalid_userdefined_options(
                    not_valid_options,
                    self._option_type,
                    "a-z A-Z 0-9 /_-"
                )
            ]
        return []

class DependsOnOption(KeyValidator):
    """
    Report REQUIRED_OPTION_IS_MISSING when the option_dict contains
    option_name_list options and does not contain the prerequisite_option
    """
    def __init__(
        self, option_name_list, prerequisite_name, option_type=None,
        prerequisite_type=None
    ):
        """
        string prerequisite_name -- name of the prerequisite options
        string prerequisite_type -- describes the prerequisite for reporting
        """
        super().__init__(option_name_list, option_type=option_type)
        self._prerequisite_name = prerequisite_name
        self._prerequisite_type = prerequisite_type

    def validate(self, option_dict):
        return [
            reports.prerequisite_option_is_missing(
                option_name,
                self._prerequisite_name,
                self._option_type,
                self._prerequisite_type
            )
            for option_name in self._option_name_list
            if (
                option_name in option_dict
                and
                self._prerequisite_name not in option_dict
            )
        ]

class IsRequiredAll(KeyValidator):
    """
    Report REQUIRED_OPTIONS_ARE_MISSING with all option_name_list options
    missing in option_dict
    """
    def validate(self, option_dict):
        missing = set(self._option_name_list) - set(option_dict.keys())
        if missing:
            return [reports.required_options_are_missing(
                missing,
                self._option_type,
            )]
        return []

class IsRequiredSome(KeyValidator):
    """
    Report REQUIRED_OPTIONS_ARE_MISSING when the option_dict does not contain
    at least one item from the option_name_list
    """
    def validate(self, option_dict):
        found = set(self._option_name_list) & set(option_dict.keys())
        if not found:
            return [reports.required_option_of_alternatives_is_missing(
                self._option_name_list,
                self._option_type,
            )]
        return []

class MutuallyExclusive(KeyValidator):
    """
    Report MUTUALLY_EXCLUSIVE_OPTIONS when there is more than one of
    mutually_exclusive_names in option_dict.
    """
    def validate(self, option_dict):
        found = set(self._option_name_list) & set(option_dict.keys())
        if len(found) > 1:
            return [reports.mutually_exclusive_options(
                found,
                self._option_type,
            )]
        return []

class NamesIn(KeyValidator):
    """
    Report INVALID_OPTIONS for option_dict keys not in option_name_list
    """
    def __init__(
        self, option_name_list, option_type=None,
        allowed_option_patterns=None, banned_name_list=None,
        code_for_warning=None, produce_warning=False,
    ):
        """
        mixed allowed_option_patterns -- option patterns to be added to a report
        list banned_name_list -- list of options which cannot be forced
        string code_for_warning -- which code makes this produce warnings
        bool produce_warning -- False produces an error, True a warning
        """
        super().__init__(option_name_list, option_type=option_type)
        self._allowed_option_patterns = allowed_option_patterns or []
        self._banned_name_set = set(banned_name_list or [])
        self._code_for_warning = code_for_warning
        self._produce_warning = produce_warning

    def validate(self, option_dict):
        name_set = set(option_dict.keys())
        banned_names = set()
        if not (self._code_for_warning is None and not self._produce_warning):
            banned_names = name_set & self._banned_name_set
        invalid_names = name_set - set(self._option_name_list) - banned_names

        report_list = []

        create_report = reports.get_problem_creator(
            self._code_for_warning,
            self._produce_warning
        )
        if invalid_names:
            report_list.append(create_report(
                reports.invalid_options,
                invalid_names,
                self._option_name_list,
                self._option_type,
                allowed_option_patterns=self._allowed_option_patterns,
            ))
        if banned_names:
            report_list.append(reports.invalid_options(
                banned_names,
                self._option_name_list,
                self._option_type,
            ))
        return report_list

### values validators

class ValueValidator(ValidatorInterface):
    def __init__(self, option_name, option_name_for_report=None):
        """
        srring option_name -- name of the option to check
        string option_name_for_report -- optional option_name override
        """
        self._option_name = option_name
        self._option_name_for_report = option_name_for_report
        self.empty_string_valid = False

    def validate(self, option_dict):
        if self._option_name not in option_dict:
            return []
        value = ValuePair.get(option_dict[self._option_name])
        if self.empty_string_valid and is_empty_string(value.normalized):
            return []
        return self._validate_value(value)

    def _get_option_name_for_report(self):
        return (
            self._option_name_for_report
                if self._option_name_for_report is not None
                else self._option_name
        )

    def _validate_value(self, value):
        raise NotImplementedError()

class ValuePredicateBase(ValueValidator):
    """
    Base class for simple predicate validators
    """
    def __init__(
        self, option_name, option_name_for_report=None,
        code_for_warning=None, produce_warning=False,
    ):
        """
        string code_for_warning -- which code makes this produce warnings
        bool produce_warning -- False produces an error, True a warning
        """
        super().__init__(
            option_name,
            option_name_for_report=option_name_for_report
        )
        self._code_for_warning = code_for_warning
        self._produce_warning = produce_warning
        self._value_cannot_be_empty = False
        self._forbidden_characters = None

    def _validate_value(self, value):
        if not self._is_valid(value.normalized):
            create_report = reports.get_problem_creator(
                self._code_for_warning,
                self._produce_warning
            )
            return [create_report(
                reports.invalid_option_value,
                self._get_option_name_for_report(),
                value.original,
                self._get_allowed_values(),
                cannot_be_empty=self._value_cannot_be_empty,
                forbidden_characters=self._forbidden_characters,
            )]

        return []

    def _is_valid(self, value):
        raise NotImplementedError()

    def _get_allowed_values(self):
        raise NotImplementedError()

class ValueCorosyncValue(ValueValidator):
    """
    Report INVALID_OPTION_VALUE when a value is not a valid corosync value

    This is meant to prevent entering characters which could be used to add
    custom sections / values to corosync.conf. It must never be allowed to
    force this validator.
    Other validators are meant to check if the value is otherwise valid
    (ValueIn, ValueIntegerInRange...) and empty / not empty. Their errors may be
    made forcible.
    """
    def _validate_value(self, value):
        if not isinstance(value.normalized, str):
            return []
        forbidden_characters = "{}\n\r"
        if set(value.normalized) & set(forbidden_characters):
            # We must be strict and do not allow to override this validation,
            # otherwise setting a cratfed option value could be misused for
            # setting arbitrary corosync.conf settings.
            return [
                reports.invalid_option_value(
                    self._get_option_name_for_report(),
                    value.original,
                    None,
                    # Make it actually print "\n" and "\r" strings instead of
                    # going to the next line.
                    # Let the user know all forbidden characters right away. Do
                    # not let them try them one by one by only reporting those
                    # actually used in the value.
                    forbidden_characters="{}\\n\\r"
                )
            ]
        return []

class ValueId(ValueValidator):
    """
    Report ID errors and optionally book IDs along the way
    """
    def __init__(
        self, option_name, option_name_for_report=None, id_provider=None
    ):
        """
        IdProvider id_provider -- checks id uniqueness and books ids if set
        """
        super().__init__(
            option_name,
            option_name_for_report=option_name_for_report
        )
        self._id_provider = id_provider

    def _validate_value(self, value):
        report_list = []
        validate_id(value.normalized, self._option_name_for_report, report_list)
        if self._id_provider is not None and not report_list:
            report_list.extend(
                self._id_provider.book_ids(value.normalized)
            )
        return report_list

class ValueIn(ValuePredicateBase):
    """
    Report INVALID_OPTION_VALUE when a value is not in a set of allowed values
    """
    def __init__(
        self, option_name, allowed_value_list, option_name_for_report=None,
        code_for_warning=None, produce_warning=False,
    ):
        """
        list of string allowed_value_list -- list of possible values
        """
        super().__init__(
            option_name,
            option_name_for_report=option_name_for_report,
            code_for_warning=code_for_warning,
            produce_warning=produce_warning,
        )
        self._allowed_value_list = allowed_value_list

    def _is_valid(self, value):
        return value in self._allowed_value_list

    def _get_allowed_values(self):
        return self._allowed_value_list

class ValueIntegerInRange(ValuePredicateBase):
    """
    Report INVALID_OPTION_VALUE when the value is not an integer such that
    at_least <= value <= at_most
    """
    def __init__(
        self, option_name, at_least, at_most, option_name_for_report=None,
        code_for_warning=None, produce_warning=False,
    ):
        """
        int at_least -- minimal allowed value
        int at_most -- maximal allowed value
        """
        super().__init__(
            option_name,
            option_name_for_report=option_name_for_report,
            code_for_warning=code_for_warning,
            produce_warning=produce_warning,
        )
        self._at_least = at_least
        self._at_most = at_most

    def _is_valid(self, value):
        return is_integer(value, self._at_least, self._at_most)

    def _get_allowed_values(self):
        return f"{self._at_least}..{self._at_most}"

class ValueIpAddress(ValuePredicateBase):
    """
    Report INVALID_OPTION_VALUE when the value is not an IP address
    """
    def _is_valid(self, value):
        return is_ipv4_address(value) or is_ipv6_address(value)

    def _get_allowed_values(self):
        return "an IP address"

class ValueNonnegativeInteger(ValuePredicateBase):
    """
    Report INVALID_OPTION_VALUE when the value is not an integer greater than -1
    """
    def _is_valid(self, value):
        return is_integer(value, 0)

    def _get_allowed_values(self):
        return "a non-negative integer"

class ValueNotEmpty(ValuePredicateBase):
    """
    Report INVALID_OPTION_VALUE when a value is empty
    """
    def __init__(
        self, option_name, value_desc_or_enum, option_name_for_report=None,
        code_for_warning=None, produce_warning=False,
    ):
        """
        mexed value_desc_or_enum -- a list or a description of possible values
        """
        super().__init__(
            option_name,
            option_name_for_report=option_name_for_report,
            code_for_warning=code_for_warning,
            produce_warning=produce_warning,
        )
        self._value_desc_or_enum = value_desc_or_enum
        self._value_cannot_be_empty = True

    def _is_valid(self, value):
        return not is_empty_string(value)

    def _get_allowed_values(self):
        return self._value_desc_or_enum

class ValuePortNumber(ValuePredicateBase):
    """
    Report INVALID_OPTION_VALUE when the value is not a TCP or UDP port number
    """
    def _is_valid(self, value):
        return is_port_number(value)

    def _get_allowed_values(self):
        return "a port number (1..65535)"

class ValuePortRange(ValuePredicateBase):
    """
    Report INVALID_OPTION_VALUE when the value is not a TCP or UDP port range
    """
    def _is_valid(self, value):
        return (
            matches_regexp(value, "^[0-9]+-[0-9]+$")
            and
            all([is_port_number(part) for part in value.split("-", 1)])
        )

    def _get_allowed_values(self):
        return "port-port"

class ValuePositiveInteger(ValuePredicateBase):
    """
    Report INVALID_OPTION_VALUE when the value is not an integer greater than 0
    """
    def _is_valid(self, value):
        return is_integer(value, 1)

    def _get_allowed_values(self):
        return "a positive integer"

class ValueTimeInterval(ValuePredicateBase):
    """
    Report INVALID_OPTION_VALUE when the value is not a time interval
    """
    def _is_valid(self, value):
        return timeout_to_seconds(value) is not None

    def _get_allowed_values(self):
        return "time interval (e.g. 1, 2s, 3m, 4h, ...)"

### predicates

def is_empty_string(value):
    """
    Check if the specified value is an empty string

    mixed value -- value to check
    """
    return isinstance(value, str) and not value

def is_integer(value, at_least=None, at_most=None):
    """
    Check if the specified value is an integer, optionally check a range

    mixed value -- string, int or float, value to check
    """
    try:
        if value is None or isinstance(value, float):
            return False
        if isinstance(value, str) and not _INTEGER_RE.fullmatch(value):
            return False
        value_int = int(value)
        if at_least is not None and value_int < at_least:
            return False
        if at_most is not None and value_int > at_most:
            return False
    except ValueError:
        return False
    return True

def is_ipv4_address(value):
    """
    Check if the specified value is an IPv4 address

    string value -- value to check
    """
    try:
        # ip_address accepts both strings and integers. We check for "." to
        # make sure it is a string representation of an IP.
        if "." in value:
            ipaddress.IPv4Address(value)
            return True
        return False
    except (TypeError, ValueError):
        # not an IP address
        return False

def is_ipv6_address(value):
    """
    Check if the specified value is an IPv6 address

    string value -- value to check
    """
    try:
        # ip_address accepts both strings and integers. We check for ":" to
        # make sure it is a string representation of an IP.
        if ":" in value:
            ipaddress.IPv6Address(value)
            return True
        return False
    except (TypeError, ValueError):
        # not an IP address
        return False

def is_port_number(value):
    """
    Check if the specified value is a TCP or UDP port number

    mixed value -- string, int or float, value to check
    """
    return is_integer(value, 1, 65535)

def matches_regexp(value, regexp):
    """
    Check if the specified value matches the specified regular expression

    mixed value -- string, int or float, value to check
    mixed regexp -- string or RegularExpression to match the value against
    """
    if not hasattr(regexp, "match"):
        regexp = re.compile(regexp)
    return regexp.match(value) is not None

### tools

def set_warning(code_for_warning, produce_warning):
    """
    A usefull shortcut for calling validators

    string code_for_warning -- code to force an error
    bool produce_warning -- report warnings instead of errors
    """
    return {
        "code_for_warning": code_for_warning,
        "produce_warning": produce_warning,
    }



import socket

from tornado.iostream import IOStream

from pcs.daemon import log

async def notify(socket_name):
    if socket_name[0] == '@':
        # abstract namespace socket
        socket_name = '\0' + socket_name[1:]

    log.pcsd.info("Notifying systemd we are running (socket '%s')", socket_name)
    try:
        stream = IOStream(socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM))
        await stream.connect(socket_name)
        await stream.write(b'READY=1')
        stream.close()
        # pylint: disable=broad-except
    except Exception as e:
        log.pcsd.error("Unable to notify systemd on '%s': %s", socket_name, e)



from functools import partial

from pcs.lib.cib.constraint import colocation
import pcs.lib.commands.constraint.common

#configure common constraint command
show = partial(
    pcs.lib.commands.constraint.common.show,
    colocation.TAG_NAME,
    lambda element: element.attrib.has_key('rsc')
)

#configure common constraint command
create_with_set = partial(
    pcs.lib.commands.constraint.common.create_with_set,
    colocation.TAG_NAME,
    colocation.prepare_options_with_set
)



import re
from functools import partial

# pylint: disable=protected-access

palete = {
    "black": '\033[30m',
    "red": '\033[31m',
    "green": '\033[32m',
    "orange": '\033[33m',
    "blue": '\033[34m',
    "purple": '\033[35m',
    "cyan": '\033[36m',
    "lightgrey": '\033[37m',
    "darkgrey": '\033[90m',
    "lightred": '\033[91m',
    "lightgreen": '\033[92m',
    "yellow": '\033[93m',
    "lightblue": '\033[94m',
    "pink": '\033[95m',
    "lightcyan": '\033[96m',
    "end" : '\033[0m',
    "bold" : '\033[1m',
    "underline" : '\033[4m',
}

separator1 = '=' * 70
separator2 = '-' * 70

#apply is builtin but is deprecated since 2.3 => no problem to redefine it here
def apply(key_list, text):
    return "".join([palete[key] for key in key_list]) + text + palete["end"]

lightgrey = partial(apply, ["lightgrey"])
bold = partial(apply, ["bold"])
blue = partial(apply, ["blue", "bold"])
red = partial(apply, ["red", "bold"])
green = partial(apply, ["green", "bold"])

def format_module_name(name):
    prefix = ""
    part_list = name.split("_")

    if part_list[0].startswith("test"):
        prefix = "test"
        part_list[0] = part_list[0][len("test"):]

    return prefix + "_".join([bold(part) for part in part_list])

def format_module(test):
    parts = test.__class__.__module__.split(".")
    return lightgrey(".").join(parts[:-1] + [format_module_name(parts[-1])])

def format_test_method_name(test):
    if not hasattr(test, "_testMethodName"):
        return test.id()

    parts = test._testMethodName.split("_")
    if parts[0].startswith("test"):
        parts[0] = lightgrey("test") + parts[0][len("test"):]
    return lightgrey("_").join(parts)

def format_error_overview(errors, failures, slash_last):
    return [
        red("for running failed tests only (errors are first then failures):"),
        "",
    ] + [
        lightgrey(err) for err in slash_errors(
            [format_test_name(test) for test, _ in errors + failures],
            slash_last
        )
    ] + [""]

def slash_errors(error_list, slash_last=True):
    if not slash_last:
        return slash_errors(error_list[:-1]) + [error_list[-1]]
    return ["{0} \\".format(err) for err in error_list]

def format_test_name(test):
    if (
        test.__class__.__module__ == "subunit"
        and
        test.__class__.__name__ == "RemotedTestCase"
    ):
        return format_test_method_name(test)
    return (
        format_module(test)
        + "." + test.__class__.__name__
        + "." + format_test_method_name(test)
    )

def get_description(test, descriptions):
    doc_first_line = test.shortDescription()
    if descriptions and doc_first_line:
        return '\n'.join((str(test), doc_first_line))
    module_parts = test.__class__.__module__.split(".")
    module = module_parts[-1]
    package = ".".join(module_parts[:-1])+"." if module_parts else ""

    return (
        test._testMethodName
        +" "
        +lightgrey("(")
        +lightgrey(package)
        +bold(module)
        +"."
        +test.__class__.__name__
        +lightgrey(")")
    )

def format_error_list(flavour, errors, descriptions, traceback_highlight):
    line_list = []
    for test, err in errors:
        line_list.extend([
            lightgrey(separator1),
            "%s: %s" % (red(flavour), get_description(test, descriptions)),
            lightgrey(separator2),
            "%s" % format_traceback(err) if traceback_highlight else err,
        ])
    return line_list

def format_traceback(err):
    formated_err = []
    path_regex = re.compile(
        r'^  File "(?P<path>[^"]+)", line (?P<line>\d+), in (?P<name>.*)$'
    )
    was_prev_path = False
    for line in err.splitlines():
        if line == "Traceback (most recent call last):":
            formated_err.append(lightgrey(line))
            was_prev_path = False
            continue

        match = path_regex.match(line)
        if match:
            path = match.group("path").split("/")
            formated_err.append(
                lightgrey('  File "')
                + lightgrey("/").join(path[:-1] + [bold(path[-1])])
                + lightgrey('", line ') + bold(match.group("line"))
                + lightgrey(', in ') + bold(match.group("name"))
            )
            was_prev_path = True
        elif was_prev_path:
            formated_err.append(bold(line))
            was_prev_path = False
        else:
            formated_err.append(line)
            was_prev_path = False
    return "\n".join(formated_err)+"\n"

def format_skips(skip_map):
    return [blue("Some tests have been skipped:")] + [
        lightgrey("{0} ({1}x)".format(reason, len(test_list)))
        for reason, test_list in skip_map.items()
    ] + [""]



from unittest import TestCase
from lxml import etree

from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_report_item_list_equal,
    assert_xml_equal,
)
from pcs_test.tools.custom_mock import MockLibraryReportProcessor
from pcs_test.tools.misc import create_patcher
from pcs_test.tools.xml import etree_to_str

from pcs.common import report_codes
from pcs.common.fencing_topology import (
    TARGET_TYPE_NODE,
    TARGET_TYPE_REGEXP,
    TARGET_TYPE_ATTRIBUTE,
)
from pcs.lib import reports
from pcs.lib.errors import LibraryError, ReportItemSeverity as severity
from pcs.lib.pacemaker.state import ClusterState

from pcs.lib.cib import fencing_topology as lib


patch_lib = create_patcher("pcs.lib.cib.fencing_topology")

# pylint: disable=too-few-public-methods, anomalous-backslash-in-string
# pylint: disable=protected-access, no-self-use


class CibMixin:
    @staticmethod
    def get_cib():
        return etree.fromstring("""
            <cib><configuration>
            <fencing-topology>
                <fencing-level
                    id="fl1" index="1" devices="d1,d2" target="nodeA"
                />
                <fencing-level
                    id="fl2" index="2" devices="d3" target="nodeA"
                />
                <fencing-level
                    id="fl3" index="1" devices="d2,d1" target="nodeB"
                />
                <fencing-level
                    id="fl4" index="2" devices="d3" target="nodeB"
                />
                <fencing-level
                    id="fl5" index="1" devices="d3,d4" target-pattern="node\\d+"
                />
                <fencing-level
                    id="fl6" index="2" devices="d1" target-pattern="node\\d+"
                />
                <fencing-level
                    id="fl7" index="3" devices="d3,d4"
                    target-attribute="fencing" target-value="improved"
                />
                <fencing-level
                    id="fl8" index="4" devices="d5"
                    target-attribute="fencing" target-value="improved"
                />
                <fencing-level
                    id="fl9" index="3" devices="dR" target-pattern="node-R.*"
                />
                <fencing-level
                    id="fl10" index="4" devices="dR-special"
                    target-attribute="fencing" target-value="remote-special"
                />
            </fencing-topology>
            </configuration></cib>
        """)


class StatusNodesMixin:
    def get_status(self):
        return ClusterState("""
            <crm_mon version="2.0.3">
                <summary>
                    <stack type="corosync" />
                    <current_dc present="true" />
                    <last_update time="Wed Nov  6 13:45:41 2019" />
                    <last_change time="Wed Nov  6 10:42:54 2019"
                        user="hacluster" client="crmd" origin="node1"
                    />
                    <nodes_configured number="2" />
                    <resources_configured number="0" disabled="0" blocked="0" />
                    <cluster_options stonith-enabled="true"
                        symmetric-cluster="true" no-quorum-policy="stop"
                        maintenance-mode="false"
                    />
                </summary>
                <nodes>
                    <node name="nodeA" id="1" online="true" standby="false"
                        standby_onfail="false" maintenance="false"
                        pending="false" unclean="false" shutdown="false"
                        expected_up="true" is_dc="true" resources_running="0"
                        type="member"
                    />
                    <node name="nodeB" id="2" online="true" standby="false"
                        standby_onfail="false" maintenance="false"
                        pending="false" unclean="false" shutdown="false"
                        expected_up="true" is_dc="false" resources_running="0"
                        type="member"
                    />
                </nodes>
            </crm_mon>
        """).node_section.nodes


@patch_lib("_append_level_element")
@patch_lib("_validate_level_target_devices_does_not_exist")
@patch_lib("_validate_devices")
@patch_lib("_validate_target")
@patch_lib("_validate_level", return_value="valid_level")
class AddLevel(TestCase):
    def setUp(self):
        self.reporter = MockLibraryReportProcessor()

    def assert_validators_called(
        self, mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl,
        dupl_called=True
    ):
        mock_val_level.assert_called_once_with(self.reporter, "level")
        mock_val_target.assert_called_once_with(
            self.reporter, "cluster_status_nodes", "target_type",
            "target_value", "force_node"
        )
        mock_val_devices.assert_called_once_with(
            self.reporter, "resources_el", "devices", "force_device"
        )
        if dupl_called:
            mock_val_dupl.assert_called_once_with(
                self.reporter, "topology_el", "level", "target_type",
                "target_value", "devices"
            )
        else:
            mock_val_dupl.assert_not_called()

    def assert_called_invalid(
        self, mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl,
        mock_append, dupl_called=True
    ):
        self.assertRaises(
            LibraryError,
            lambda: lib.add_level(
                self.reporter, "topology_el", "resources_el", "level",
                "target_type", "target_value", "devices",
                "cluster_status_nodes", "force_device", "force_node"
            )
        )
        self.assert_validators_called(
            mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl,
            dupl_called
        )
        mock_append.assert_not_called()

    def test_success(
        self, mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl,
        mock_append
    ):
        lib.add_level(
            self.reporter, "topology_el", "resources_el", "level",
            "target_type", "target_value", "devices", "cluster_status_nodes",
            "force_device", "force_node"
        )
        self.assert_validators_called(
            mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl
        )
        mock_append.assert_called_once_with(
            "topology_el", "valid_level", "target_type", "target_value",
            "devices"
        )

    def test_invalid_level(
        self, mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl,
        mock_append
    ):
        mock_val_level.side_effect = lambda reporter, level: reporter.append(
            reports.invalid_option_value("level", level, "a positive integer")
        )
        self.assert_called_invalid(
            mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl,
            mock_append, dupl_called=False
        )

    def test_invalid_target(
        self, mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl,
        mock_append
    ):
        mock_val_target.side_effect = (
            lambda reporter, status_nodes, target_type, target_value, force:
                reporter.append(
                    reports.node_not_found(target_value)
                )
        )
        self.assert_called_invalid(
            mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl,
            mock_append, dupl_called=False
        )

    def test_invalid_devices(
        self, mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl,
        mock_append
    ):
        mock_val_devices.side_effect = (
            lambda reporter, resources, devices, force:
                reporter.append(
                    reports.stonith_resources_do_not_exist(["device"])
                )
        )
        self.assert_called_invalid(
            mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl,
            mock_append, dupl_called=False
        )

    def test_already_exists(
        self, mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl,
        mock_append
    ):
        mock_val_dupl.side_effect = (
            lambda reporter, tree, level, target_type, target_value, devices:
                reporter.append(
                    reports.fencing_level_already_exists(
                        level, target_type, target_value, devices
                    )
                )
        )
        self.assert_called_invalid(
            mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl,
            mock_append, dupl_called=True
        )


class RemoveAllLevels(TestCase, CibMixin):
    def setUp(self):
        self.cib = self.get_cib()
        self.tree = self.cib.find("configuration/fencing-topology")

    def test_success(self):
        lib.remove_all_levels(self.tree)
        assert_xml_equal(
            "<fencing-topology />",
            etree_to_str(self.tree)
        )


class RemoveLevelsByParams(TestCase, CibMixin):
    def setUp(self):
        self.cib = self.get_cib()
        self.tree = self.cib.find("configuration/fencing-topology")
        self.reporter = MockLibraryReportProcessor()

    def get_remaining_ids(self):
        self.assertTrue(
            self.cib.find("configuration/fencing-topology") is not None
        )
        return [el.get("id") for el in self.tree.findall("fencing-level")]

    def test_level(self):
        lib.remove_levels_by_params(
            self.reporter, self.tree, level=2
        )
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl1", "fl3", "fl5", "fl7", "fl8", "fl9", "fl10"]
        )
        assert_report_item_list_equal(self.reporter.report_item_list, [])

    def test_target_node(self):
        lib.remove_levels_by_params(
            self.reporter, self.tree, target_type=TARGET_TYPE_NODE,
            target_value="nodeA"
        )
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl3", "fl4", "fl5", "fl6", "fl7", "fl8", "fl9", "fl10"]
        )
        assert_report_item_list_equal(self.reporter.report_item_list, [])

    def test_target_pattern(self):
        lib.remove_levels_by_params(
            self.reporter, self.tree, target_type=TARGET_TYPE_REGEXP,
            target_value=r"node\d+"
        )
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl1", "fl2", "fl3", "fl4", "fl7", "fl8", "fl9", "fl10"]
        )
        assert_report_item_list_equal(self.reporter.report_item_list, [])

    def test_target_attrib(self):
        lib.remove_levels_by_params(
            self.reporter, self.tree, target_type=TARGET_TYPE_ATTRIBUTE,
            target_value=("fencing", "improved")
        )
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl1", "fl2", "fl3", "fl4", "fl5", "fl6", "fl9", "fl10"]
        )
        assert_report_item_list_equal(self.reporter.report_item_list, [])

    def test_one_device(self):
        lib.remove_levels_by_params(
            self.reporter, self.tree, devices=["d3"]
        )
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl1", "fl3", "fl5", "fl6", "fl7", "fl8", "fl9", "fl10"]
        )
        assert_report_item_list_equal(self.reporter.report_item_list, [])

    def test_more_devices(self):
        lib.remove_levels_by_params(
            self.reporter, self.tree, devices=["d2", "d1"]
        )
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl1", "fl2", "fl4", "fl5", "fl6", "fl7", "fl8", "fl9", "fl10"]
        )
        assert_report_item_list_equal(self.reporter.report_item_list, [])

    def test_combination(self):
        lib.remove_levels_by_params(
            self.reporter, self.tree, 2, TARGET_TYPE_NODE, "nodeB", ["d3"]
        )
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl1", "fl2", "fl3", "fl5", "fl6", "fl7", "fl8", "fl9", "fl10"]
        )
        assert_report_item_list_equal(self.reporter.report_item_list, [])

    def test_invalid_target(self):
        assert_raise_library_error(
            lambda: lib.remove_levels_by_params(
                self.reporter, self.tree, target_type="bad_target",
                target_value="nodeA"
            ),
            (
                severity.ERROR,
                report_codes.INVALID_OPTION_TYPE,
                {
                    "option_name": "target",
                    "allowed_types": [
                        "node",
                        "regular expression",
                        "attribute_name=value"
                    ]
                },
                None
            ),
        )
        self.assertEqual(
            self.get_remaining_ids(),
            [
                "fl1", "fl2", "fl3", "fl4", "fl5", "fl6", "fl7", "fl8", "fl9",
                "fl10"
            ]
        )

    def test_no_such_level(self):
        assert_raise_library_error(
            lambda: lib.remove_levels_by_params(
                self.reporter, self.tree, 9, TARGET_TYPE_NODE, "nodeB", ["d3"]
            ),
            (
                severity.ERROR,
                report_codes.CIB_FENCING_LEVEL_DOES_NOT_EXIST,
                {
                    "devices": ["d3", ],
                    "target_type": TARGET_TYPE_NODE,
                    "target_value": "nodeB",
                    "level": 9,
                },
                None
            ),
        )
        self.assertEqual(
            self.get_remaining_ids(),
            [
                "fl1", "fl2", "fl3", "fl4", "fl5", "fl6", "fl7", "fl8", "fl9",
                "fl10"
            ]
        )

    def test_no_such_level_ignore_missing(self):
        lib.remove_levels_by_params(
            self.reporter, self.tree, 9, TARGET_TYPE_NODE, "nodeB", ["d3"], True
        )
        self.assertEqual(
            self.get_remaining_ids(),
            [
                "fl1", "fl2", "fl3", "fl4", "fl5", "fl6", "fl7", "fl8", "fl9",
                "fl10"
            ]
        )

class RemoveDeviceFromAllLevels(TestCase, CibMixin):
    def setUp(self):
        self.cib = self.get_cib()
        self.tree = self.cib.find("configuration/fencing-topology")

    def test_success(self):
        lib.remove_device_from_all_levels(self.tree, "d3")
        assert_xml_equal(
            """
            <fencing-topology>
                <fencing-level
                    id="fl1" index="1" devices="d1,d2" target="nodeA"
                />
                <fencing-level
                    id="fl3" index="1" devices="d2,d1" target="nodeB"
                />
                <fencing-level
                    id="fl5" index="1" devices="d4" target-pattern="node\\d+"
                />
                <fencing-level
                    id="fl6" index="2" devices="d1" target-pattern="node\\d+"
                />
                <fencing-level
                    id="fl7" index="3" devices="d4"
                    target-attribute="fencing" target-value="improved"
                />
                <fencing-level
                    id="fl8" index="4" devices="d5"
                    target-attribute="fencing" target-value="improved"
                />
                <fencing-level
                    id="fl9" index="3" devices="dR" target-pattern="node-R.*"
                />
                <fencing-level
                    id="fl10" index="4" devices="dR-special"
                    target-attribute="fencing" target-value="remote-special"
                />
            </fencing-topology>
            """,
            etree_to_str(self.tree)
        )

    def test_no_such_device(self):
        original_xml = etree_to_str(self.tree)
        lib.remove_device_from_all_levels(self.tree, "dX")
        assert_xml_equal(original_xml, etree_to_str(self.tree))


class Export(TestCase, CibMixin):
    def setUp(self):
        self.cib = self.get_cib()
        self.tree = self.cib.find("configuration/fencing-topology")

    def test_empty(self):
        self.assertEqual(
            lib.export(etree.fromstring("<fencing-topology />")),
            []
        )

    def test_success(self):
        self.assertEqual(
            lib.export(self.tree),
            [
                {
                    "level": "1",
                    "target_type": "node",
                    "target_value": "nodeA",
                    "devices": ["d1", "d2"],
                },
                {
                    "level": "2",
                    "target_type": "node",
                    "target_value": "nodeA",
                    "devices": ["d3"],
                },
                {
                    "level": "1",
                    "target_type": "node",
                    "target_value": "nodeB",
                    "devices": ["d2", "d1"],
                },
                {
                    "level": "2",
                    "target_type": "node",
                    "target_value": "nodeB",
                    "devices": ["d3"],
                },
                {
                    "level": "1",
                    "target_type": "regexp",
                    "target_value": r"node\d+",
                    "devices": ["d3", "d4"],
                },
                {
                    "level": "2",
                    "target_type": "regexp",
                    "target_value": r"node\d+",
                    "devices": ["d1"],
                },
                {
                    "level": "3",
                    "target_type": "attribute",
                    "target_value": ("fencing", "improved"),
                    "devices": ["d3", "d4"],
                },
                {
                    "level": "4",
                    "target_type": "attribute",
                    "target_value": ("fencing", "improved"),
                    "devices": ["d5"],
                },
                {
                    "level": "3",
                    "target_type": "regexp",
                    "target_value": "node-R.*",
                    "devices": ["dR"],
                },
                {
                    "level": "4",
                    "target_type": "attribute",
                    "target_value": ("fencing", "remote-special"),
                    "devices": ["dR-special"],
                }
            ]
        )


class Verify(TestCase, CibMixin, StatusNodesMixin):
    def setUp(self):
        self.cib = self.get_cib()
        self.tree = self.cib.find("configuration/fencing-topology")

    def fixture_resource(self, tree, name):
        el = etree.SubElement(tree, "primitive", id=name, type="fence_dummy")
        el.set("class", "stonith")

    def test_empty(self):
        resources = etree.fromstring("<resources />")
        topology = etree.fromstring("<fencing-topology />")
        reporter = MockLibraryReportProcessor()

        lib.verify(reporter, topology, resources, self.get_status())

        assert_report_item_list_equal(reporter.report_item_list, [])

    def test_success(self):
        resources = etree.fromstring("<resources />")
        for name in ["d1", "d2", "d3", "d4", "d5", "dR", "dR-special"]:
            self.fixture_resource(resources, name)
        reporter = MockLibraryReportProcessor()

        lib.verify(reporter, self.tree, resources, self.get_status())

        assert_report_item_list_equal(reporter.report_item_list, [])

    def test_failures(self):
        resources = etree.fromstring("<resources />")
        reporter = MockLibraryReportProcessor()

        lib.verify(reporter, self.tree, resources, [])

        report = [
            (
                severity.ERROR,
                report_codes.STONITH_RESOURCES_DO_NOT_EXIST,
                {
                    "stonith_ids": [
                        "d1", "d2", "d3", "d4", "d5", "dR", "dR-special"
                    ],
                },
                None
            ),
            (
                severity.ERROR,
                report_codes.NODE_NOT_FOUND,
                {
                    "node": "nodeA",
                },
                None
            ),
            (
                severity.ERROR,
                report_codes.NODE_NOT_FOUND,
                {
                    "node": "nodeB",
                },
                None
            ),
        ]
        assert_report_item_list_equal(reporter.report_item_list, report)


class ValidateLevel(TestCase):
    def test_success(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_level(reporter, 1)
        lib._validate_level(reporter, "1")
        lib._validate_level(reporter, 9)
        lib._validate_level(reporter, "9")
        lib._validate_level(reporter, "05")
        assert_report_item_list_equal(reporter.report_item_list, [])

    def test_invalid(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_level(reporter, "")
        lib._validate_level(reporter, 0)
        lib._validate_level(reporter, "0")
        lib._validate_level(reporter, -1)
        lib._validate_level(reporter, "-1")
        lib._validate_level(reporter, "1abc")
        report_list = []
        for value in ["", 0, "0", -1, "-1", "1abc"]:
            report_list.append((
                severity.ERROR,
                report_codes.INVALID_OPTION_VALUE,
                {
                    "option_value": value,
                    "option_name": "level",
                    "allowed_values": "a positive integer",
                    "cannot_be_empty": False,
                    "forbidden_characters": None,
                },
                None
            ))
        assert_report_item_list_equal(reporter.report_item_list, report_list)


@patch_lib("_validate_target_valuewise")
@patch_lib("_validate_target_typewise")
class ValidateTarget(TestCase):
    def test_delegate(self, validate_type, validate_value):
        lib._validate_target("reporter", "status", "type", "value", "force")
        validate_type.assert_called_once_with("reporter", "type")
        validate_value.assert_called_once_with(
            "reporter", "status", "type", "value", "force"
        )


class ValidateTargetTypewise(TestCase):
    def test_success(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_target_typewise(reporter, TARGET_TYPE_NODE)
        lib._validate_target_typewise(reporter, TARGET_TYPE_ATTRIBUTE)
        lib._validate_target_typewise(reporter, TARGET_TYPE_REGEXP)
        assert_report_item_list_equal(reporter.report_item_list, [])

    def test_empty(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_target_typewise(reporter, "")
        report = [(
            severity.ERROR,
            report_codes.INVALID_OPTION_TYPE,
            {
                "option_name": "target",
                "allowed_types": [
                    "node",
                    "regular expression",
                    "attribute_name=value"
                ],
            },
            None
        )]
        assert_report_item_list_equal(reporter.report_item_list, report)

    def test_invalid(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_target_typewise(reporter, "bad_target")
        report = [(
            severity.ERROR,
            report_codes.INVALID_OPTION_TYPE,
            {
                "option_name": "target",
                "allowed_types": [
                    "node",
                    "regular expression",
                    "attribute_name=value"
                ],
            },
            None
        )]
        assert_report_item_list_equal(reporter.report_item_list, report)


class ValidateTargetValuewise(TestCase, StatusNodesMixin):
    def setUp(self):
        self.state = self.get_status()

    def test_node_valid(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_target_valuewise(
            reporter, self.state, TARGET_TYPE_NODE, "nodeA"
        )
        assert_report_item_list_equal(reporter.report_item_list, [])

    def test_node_empty(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_target_valuewise(
            reporter, self.state, TARGET_TYPE_NODE, ""
        )
        report = [(
            severity.ERROR,
            report_codes.NODE_NOT_FOUND,
            {
                "node": "",
            },
            report_codes.FORCE_NODE_DOES_NOT_EXIST
        )]
        assert_report_item_list_equal(reporter.report_item_list, report)

    def test_node_invalid(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_target_valuewise(
            reporter, self.state, TARGET_TYPE_NODE, "rh7-x"
        )
        report = [(
            severity.ERROR,
            report_codes.NODE_NOT_FOUND,
            {
                "node": "rh7-x",
            },
            report_codes.FORCE_NODE_DOES_NOT_EXIST
        )]
        assert_report_item_list_equal(reporter.report_item_list, report)

    def test_node_invalid_force(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_target_valuewise(
            reporter, self.state, TARGET_TYPE_NODE, "rh7-x", force_node=True
        )
        report = [(
            severity.WARNING,
            report_codes.NODE_NOT_FOUND,
            {
                "node": "rh7-x",
            },
            None
        )]
        assert_report_item_list_equal(reporter.report_item_list, report)

    def test_node_invalid_not_forceable(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_target_valuewise(
            reporter, self.state, TARGET_TYPE_NODE, "rh7-x", allow_force=False
        )
        report = [(
            severity.ERROR,
            report_codes.NODE_NOT_FOUND,
            {
                "node": "rh7-x",
            },
            None
        )]
        assert_report_item_list_equal(reporter.report_item_list, report)


class ValidateDevices(TestCase):
    def setUp(self):
        self.resources_el = etree.fromstring("""
            <resources>
                <primitive id="dummy"
                    class="ocf" provider="pacemaker" type="Stateful"
                 />
                <primitive id="stonith1" class="stonith" type="fence_xvm" />
                <primitive id="stonith2" class="stonith" type="fence_apc" />
            </resources>
        """)

    def test_success(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_devices(
            reporter, self.resources_el, ["stonith1"]
        )
        lib._validate_devices(
            reporter, self.resources_el, ["stonith1", "stonith2"]
        )
        assert_report_item_list_equal(reporter.report_item_list, [])

    def test_empty(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_devices(reporter, self.resources_el, [])
        report = [(
            severity.ERROR,
            report_codes.REQUIRED_OPTIONS_ARE_MISSING,
            {
                "option_type": None,
                "option_names": ["stonith devices"],
            },
            None
        )]
        assert_report_item_list_equal(reporter.report_item_list, report)

    def test_invalid(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_devices(reporter, self.resources_el, ["dummy", "fenceX"])
        report = [(
            severity.ERROR,
            report_codes.STONITH_RESOURCES_DO_NOT_EXIST,
            {
                "stonith_ids": ["dummy", "fenceX"],
            },
            report_codes.FORCE_STONITH_RESOURCE_DOES_NOT_EXIST
        )]
        assert_report_item_list_equal(reporter.report_item_list, report)

    def test_invalid_forced(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_devices(
            reporter, self.resources_el, ["dummy", "fenceX"], force_device=True
        )
        report = [(
            severity.WARNING,
            report_codes.STONITH_RESOURCES_DO_NOT_EXIST,
            {
                "stonith_ids": ["dummy", "fenceX"],
            },
            None
        )]
        assert_report_item_list_equal(reporter.report_item_list, report)

    def test_node_invalid_not_forceable(self):
        reporter = MockLibraryReportProcessor()
        lib._validate_devices(
            reporter, self.resources_el, ["dummy", "fenceX"], allow_force=False
        )
        report = [(
            severity.ERROR,
            report_codes.STONITH_RESOURCES_DO_NOT_EXIST,
            {
                "stonith_ids": ["dummy", "fenceX"],
            },
            None
        )]
        assert_report_item_list_equal(reporter.report_item_list, report)


@patch_lib("_find_level_elements")
class ValidateLevelTargetDevicesDoesNotExist(TestCase):
    def test_success(self, mock_find):
        mock_find.return_value = []
        reporter = MockLibraryReportProcessor()

        lib._validate_level_target_devices_does_not_exist(
            reporter, "tree", "level", "target_type", "target_value",
            ["devices"]
        )

        mock_find.assert_called_once_with(
            "tree", "level", "target_type", "target_value", ["devices"]
        )
        assert_report_item_list_equal(reporter.report_item_list, [])

    def test_error(self, mock_find):
        mock_find.return_value = ["element"]
        reporter = MockLibraryReportProcessor()

        lib._validate_level_target_devices_does_not_exist(
            reporter, "tree", "level", "target_type", "target_value",
            ["devices"]
        )

        mock_find.assert_called_once_with(
            "tree", "level", "target_type", "target_value", ["devices"]
        )
        report = [(
            severity.ERROR,
            report_codes.CIB_FENCING_LEVEL_ALREADY_EXISTS,
            {
                "devices": ["devices"],
                "target_type": "target_type",
                "target_value": "target_value",
                "level": "level",
            },
            None
        )]
        assert_report_item_list_equal(reporter.report_item_list, report)


class AppendLevelElement(TestCase):
    def setUp(self):
        self.tree = etree.fromstring("<fencing-topology />")

    def test_node_name(self):
        lib._append_level_element(
            self.tree, 1, TARGET_TYPE_NODE, "node1", ["d1"]
        )
        assert_xml_equal(
            """
            <fencing-topology>
                <fencing-level
                    id="fl-node1-1"
                    devices="d1" index="1" target="node1"
                />
            </fencing-topology>
            """,
            etree_to_str(self.tree)
        )

    def test_node_pattern(self):
        lib._append_level_element(
            self.tree, "2", TARGET_TYPE_REGEXP, r"node-\d+", ["d1", "d2"]
        )
        assert_xml_equal(
            """
            <fencing-topology>
                <fencing-level
                    id="fl-node-d-2"
                    devices="d1,d2" index="2" target-pattern="node-\\d+"
                />
            </fencing-topology>
            """,
            etree_to_str(self.tree)
        )

    def test_node_attribute(self):
        lib._append_level_element(
            self.tree, 3, TARGET_TYPE_ATTRIBUTE, ("name%@x", "val%@x"), ["d1"],
        )
        assert_xml_equal(
            """
            <fencing-topology>
                <fencing-level
                    id="fl-namex-3"
                    devices="d1"
                    index="3"
                    target-attribute="name%@x" target-value="val%@x"
                />
            </fencing-topology>
            """,
            etree_to_str(self.tree)
        )


class FindLevelElements(TestCase, CibMixin):
    def setUp(self):
        self.cib = self.get_cib()
        self.tree = self.cib.find("configuration/fencing-topology")

    def get_ids(self, elements):
        return [el.get("id") for el in elements]

    def test_no_filter(self):
        self.assertEqual(
            self.get_ids(lib._find_level_elements(self.tree)),
            [
                "fl1", "fl2", "fl3", "fl4", "fl5", "fl6", "fl7", "fl8", "fl9",
                "fl10"
            ]
        )

    def test_no_such_level(self):
        self.assertEqual(
            self.get_ids(lib._find_level_elements(
                self.tree, level=2, target_type=TARGET_TYPE_NODE,
                target_value="nodeB", devices=["d5"]
            )),
            []
        )

    def test_level(self):
        self.assertEqual(
            self.get_ids(lib._find_level_elements(
                self.tree, level=1
            )),
            ["fl1", "fl3", "fl5"]
        )

    def test_target_node(self):
        self.assertEqual(
            self.get_ids(lib._find_level_elements(
                self.tree, target_type=TARGET_TYPE_NODE, target_value="nodeB"
            )),
            ["fl3", "fl4"]
        )

    def test_target_pattern(self):
        self.assertEqual(
            self.get_ids(lib._find_level_elements(
                self.tree, target_type=TARGET_TYPE_REGEXP,
                target_value="node-R.*"
            )),
            ["fl9"]
        )

    def test_target_attribute(self):
        self.assertEqual(
            self.get_ids(lib._find_level_elements(
                self.tree, target_type=TARGET_TYPE_ATTRIBUTE,
                target_value=("fencing", "improved")
            )),
            ["fl7", "fl8"]
        )

    def test_devices(self):
        self.assertEqual(
            self.get_ids(lib._find_level_elements(
                self.tree, devices=["d3"]
            )),
            ["fl2", "fl4"]
        )

        self.assertEqual(
            self.get_ids(lib._find_level_elements(
                self.tree, devices=["d1", "d2"]
            )),
            ["fl1"]
        )

    def test_combination(self):
        self.assertEqual(
            self.get_ids(lib._find_level_elements(
                self.tree, 2, TARGET_TYPE_NODE, "nodeB", ["d3"]
            )),
            ["fl4"]
        )



DEFAULT_INSTANCE_NAME = "booth"

GLOBAL_KEYS = (
    "transport",
    "port",
    "name",
    "authfile",
    "maxtimeskew",
    "site",
    "arbitrator",
    "site-user",
    "site-group",
    "arbitrator-user",
    "arbitrator-group",
    "debug",
    "ticket",
)

TICKET_KEYS = (
    "acquire-after",
    "attr-prereq",
    "before-acquire-handler",
    "expire",
    "renewal-freq",
    "retries",
    "timeout",
    "weights",
)



from textwrap import dedent
from unittest import mock, TestCase

from pcs import settings
from pcs.common import file_type_codes, report_codes
from pcs.lib.commands import status
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import read_test_resource as rc_read

crm_mon_rng_with_history = rc_read("crm_mon.rng.with_fence_history.xml")
crm_mon_rng_without_history = rc_read("crm_mon.rng.without_fence_history.xml")

class FullClusterStatusPlaintext(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.node_name_list = ["node1", "node2", "node3"]
        self.maxDiff = None

    def _fixture_config_live_minimal(self):
        (self.config
            .runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .corosync_conf.load()
            .runner.cib.load(resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """)
            .runner.systemctl.is_active(
                "sbd", is_active=False, name="runner.systemctl.is_active.sbd"
            )
        )

    def _fixture_config_local_daemons(
        self,
        corosync_enabled=True, corosync_active=True,
        pacemaker_enabled=True, pacemaker_active=True,
        pacemaker_remote_enabled=False, pacemaker_remote_active=False,
        pcsd_enabled=True, pcsd_active=True,
        sbd_enabled=False, sbd_active=False,
    ):
        # pylint: disable=too-many-arguments
        (self.config
            .runner.systemctl.is_enabled(
                "corosync",
                name="runner.systemctl.is_enabled.corosync",
                is_enabled=corosync_enabled,
            )
            .runner.systemctl.is_active(
                "corosync",
                name="runner.systemctl.is_active.corosync",
                is_active=corosync_active,
            )
            .runner.systemctl.is_enabled(
                "pacemaker",
                name="runner.systemctl.is_enabled.pacemaker",
                is_enabled=pacemaker_enabled,
            )
            .runner.systemctl.is_active(
                "pacemaker",
                name="runner.systemctl.is_active.pacemaker",
                is_active=pacemaker_active,
            )
            .runner.systemctl.is_enabled(
                "pacemaker_remote",
                name="runner.systemctl.is_enabled.pacemaker_remote",
                is_enabled=pacemaker_remote_enabled,
            )
            .runner.systemctl.is_active(
                "pacemaker_remote",
                name="runner.systemctl.is_active.pacemaker_remote",
                is_active=pacemaker_remote_active,
            )
            .runner.systemctl.is_enabled(
                "pcsd",
                name="runner.systemctl.is_enabled.pcsd",
                is_enabled=pcsd_enabled,
            )
            .runner.systemctl.is_active(
                "pcsd",
                name="runner.systemctl.is_active.pcsd",
                is_active=pcsd_active,
            )
            .runner.systemctl.is_enabled(
                "sbd",
                name="runner.systemctl.is_enabled.sbd_2",
                is_enabled=sbd_enabled,
            )
            .runner.systemctl.is_active(
                "sbd",
                name="runner.systemctl.is_active.sbd_2",
                is_active=sbd_active,
            )
        )

    def test_life_cib_mocked_corosync(self):
        self.config.env.set_corosync_conf_data("corosync conf data")
        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            [
                fixture.error(
                    report_codes.LIVE_ENVIRONMENT_NOT_CONSISTENT,
                    mocked_files=[file_type_codes.COROSYNC_CONF],
                    required_files=[file_type_codes.CIB],
                ),
            ],
            expected_in_processor=False
        )

    def test_mocked_cib_life_corosync(self):
        self.config.env.set_cib_data("<cib/>")
        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            [
                fixture.error(
                    report_codes.LIVE_ENVIRONMENT_NOT_CONSISTENT,
                    mocked_files=[file_type_codes.CIB],
                    required_files=[file_type_codes.COROSYNC_CONF],
                ),
            ],
            expected_in_processor=False
        )

    def test_fail_getting_cluster_status(self):
        (self.config
            .runner.pcmk.load_state_plaintext(
                stdout="some stdout",
                stderr="some stderr",
                returncode=1,
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            [
                fixture.error(
                    report_codes.CRM_MON_ERROR,
                    reason="some stderr\nsome stdout",
                ),
            ],
            expected_in_processor=False
        )

    def test_fail_getting_corosync_conf(self):
        (self.config
            .runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .corosync_conf.load_content("invalid corosync conf")
        )
        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    report_codes.PARSE_ERROR_COROSYNC_CONF_LINE_IS_NOT_SECTION_NOR_KEY_VALUE,
                ),
            ],
            expected_in_processor=False
        )

    def test_fail_getting_cib(self):
        (self.config
            .runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .corosync_conf.load()
            .runner.cib.load_content(
                "some stdout", stderr="cib load error", returncode=1
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            [
                fixture.error(
                    report_codes.CIB_LOAD_ERROR,
                    reason="cib load error",
                ),
            ],
            expected_in_processor=False
        )

    def test_success_live(self):
        self._fixture_config_live_minimal()
        self._fixture_config_local_daemons()
        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            dedent("""\
                Cluster name: test99
                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            )
        )

    def test_success_live_verbose(self):
        (self.config
            .env.set_known_nodes(self.node_name_list)
            .fs.open(
                settings.crm_mon_schema,
                mock.mock_open(read_data=crm_mon_rng_without_history)(),
                name="fs.open.crm_mon_rng"
            )
            .runner.pcmk.load_state_plaintext(
                verbose=True,
                stdout="crm_mon cluster status",
            )
            .corosync_conf.load(node_name_list=self.node_name_list)
            .runner.cib.load(resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """)
            .runner.pcmk.load_ticket_state_plaintext(stdout="ticket status")
            .runner.systemctl.is_active(
                "sbd", is_active=False, name="runner.systemctl.is_active.sbd"
            )
        )
        self._fixture_config_local_daemons()
        (self.config
            .http.host.check_reachability(node_labels=self.node_name_list)
        )

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(),
                verbose=True
            ),
            dedent("""\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  ticket status

                PCSD Status:
                  node1: Online
                  node2: Online
                  node3: Online

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            )
        )

    def test_succes_mocked(self):
        (self.config
            .env.set_corosync_conf_data(rc_read("corosync.conf"))
            .env.set_cib_data("<cib/>")
            .runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .runner.cib.load(resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """)
        )
        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            dedent("""\
                Cluster name: test99
                crm_mon cluster status"""
            )
        )

    def test_succes_mocked_verbose(self):
        (self.config
            .env.set_corosync_conf_data(rc_read("corosync.conf"))
            .env.set_cib_data("<cib/>")
            .fs.open(
                settings.crm_mon_schema,
                mock.mock_open(read_data=crm_mon_rng_without_history)(),
                name="fs.open.crm_mon_rng"
            )
            .runner.pcmk.load_state_plaintext(
                verbose=True, stdout="crm_mon cluster status",
            )
            .runner.cib.load(resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """)
            .runner.pcmk.load_ticket_state_plaintext(stdout="ticket status")
        )
        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(),
                verbose=True
            ),
            dedent("""\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  ticket status"""
            )
        )

    def test_success_verbose_inactive_and_fence_history(self):
        (self.config
            .env.set_known_nodes(self.node_name_list)
            .fs.open(
                settings.crm_mon_schema,
                mock.mock_open(read_data=crm_mon_rng_with_history)(),
                name="fs.open.crm_mon_rng"
            )
            .runner.pcmk.load_state_plaintext(
                verbose=True,
                inactive=False,
                fence_history=True,
                stdout="crm_mon cluster status",
            )
            .corosync_conf.load(node_name_list=self.node_name_list)
            .runner.cib.load(resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """)
            .runner.pcmk.load_ticket_state_plaintext(stdout="ticket status")
            .runner.systemctl.is_active(
                "sbd", is_active=False, name="runner.systemctl.is_active.sbd"
            )
        )
        self._fixture_config_local_daemons()
        (self.config
            .http.host.check_reachability(node_labels=self.node_name_list)
        )

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(),
                verbose=True,
                hide_inactive_resources=True,
            ),
            dedent("""\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  ticket status

                PCSD Status:
                  node1: Online
                  node2: Online
                  node3: Online

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            )
        )

    def _assert_success_with_ticket_status_failure(self, stderr="", msg=""):
        (self.config
            .env.set_known_nodes(self.node_name_list)
            .fs.open(
                settings.crm_mon_schema,
                mock.mock_open(read_data=crm_mon_rng_without_history)(),
                name="fs.open.crm_mon_rng"
            )
            .runner.pcmk.load_state_plaintext(
                verbose=True,
                stdout="crm_mon cluster status",
            )
            .corosync_conf.load(node_name_list=self.node_name_list)
            .runner.cib.load(resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """)
            .runner.pcmk.load_ticket_state_plaintext(
                stdout="ticket stdout", stderr=stderr, returncode=1
            )
            .runner.systemctl.is_active(
                "sbd", is_active=False, name="runner.systemctl.is_active.sbd"
            )
        )
        self._fixture_config_local_daemons()
        (self.config
            .http.host.check_reachability(node_labels=self.node_name_list)
        )

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(),
                verbose=True
            ),
            dedent("""\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  WARNING: Unable to get information about tickets{msg}

                PCSD Status:
                  node1: Online
                  node2: Online
                  node3: Online

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ).format(msg=msg)
        )

    def test_success_with_ticket_status_failure(self):
        self._assert_success_with_ticket_status_failure()

    def test_success_with_ticket_status_failure_with_message(self):
        self._assert_success_with_ticket_status_failure(
            stderr="ticket status error\nmultiline\n",
            msg="\n    ticket status error\n    multiline",
        )

    def test_stonith_warning_no_devices(self):
        (self.config
            .runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .corosync_conf.load()
            .runner.cib.load()
            .runner.systemctl.is_active(
                "sbd", is_active=False, name="runner.systemctl.is_active.sbd"
            )
        )
        self._fixture_config_local_daemons()

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            dedent("""\
                Cluster name: test99

                WARNINGS:
                No stonith devices and stonith-enabled is not false

                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            )
        )

    def test_stonith_warning_no_devices_sbd_enabled(self):
        (self.config
            .runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .corosync_conf.load()
            .runner.cib.load()
            .runner.systemctl.is_active(
                "sbd", is_active=True, name="runner.systemctl.is_active.sbd"
            )
        )
        self._fixture_config_local_daemons()

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            dedent("""\
                Cluster name: test99
                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            )
        )

    def test_stonith_warnings_regarding_devices_configuration(self):
        (self.config
            .runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .corosync_conf.load()
            .runner.cib.load(resources="""
                <resources>
                    <primitive id="S1" class="stonith" type="fence_dummy">
                        <instance_attributes>
                            <nvpair name="method" value="cycle" />
                        </instance_attributes>
                    </primitive>
                    <primitive id="S2" class="stonith" type="fence_dummy">
                        <instance_attributes>
                            <nvpair name="action" value="value" />
                            <nvpair name="method" value="not-cycle" />
                        </instance_attributes>
                    </primitive>
                    <primitive id="S3" class="stonith" type="fence_dummy">
                        <instance_attributes>
                            <nvpair name="method" value="cycle" />
                            <nvpair name="action" value="value" />
                        </instance_attributes>
                    </primitive>
                </resources>
            """)
            .runner.systemctl.is_active(
                "sbd", is_active=False, name="runner.systemctl.is_active.sbd"
            )
        )
        self._fixture_config_local_daemons()

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            dedent("""\
                Cluster name: test99

                WARNINGS:
                Following stonith devices have the 'action' option set, it is recommended to set 'pcmk_off_action', 'pcmk_reboot_action' instead: 'S2', 'S3'
                Following stonith devices have the 'method' option set to 'cycle' which is potentially dangerous, please consider using 'onoff': 'S1', 'S3'

                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            )
        )

    def test_pcsd_status_issues(self):
        self.node_name_list = ["node1", "node2", "node3", "node4", "node5"]

        (self.config
            .env.set_known_nodes(self.node_name_list[1:])
            .fs.open(
                settings.crm_mon_schema,
                mock.mock_open(read_data=crm_mon_rng_without_history)(),
                name="fs.open.crm_mon_rng"
            )
            .runner.pcmk.load_state_plaintext(
                verbose=True,
                stdout="crm_mon cluster status",
            )
            .corosync_conf.load(node_name_list=self.node_name_list)
            .runner.cib.load(resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """)
            .runner.pcmk.load_ticket_state_plaintext(stdout="ticket status")
            .runner.systemctl.is_active(
                "sbd", is_active=False, name="runner.systemctl.is_active.sbd"
            )
        )
        self._fixture_config_local_daemons()
        (self.config
            .http.host.check_reachability(communication_list=[
                # node1 has no record in known-hosts
                dict(
                    label="node2",
                    was_connected=False,
                    errno=7,
                    error_msg="node2 error"
                ),
                dict(
                    label="node3",
                    response_code=401,
                    output="node3 output",
                ),
                dict(
                    label="node4",
                    response_code=500,
                    output="node4 output",
                ),
                dict(
                    label="node5",
                )
            ])
        )

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(),
                verbose=True
            ),
            dedent("""\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  ticket status

                PCSD Status:
                  node1: Unable to authenticate
                  node2: Offline
                  node3: Unable to authenticate
                  node4: Online
                  node5: Online

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            )
        )

    def test_daemon_status_all_on(self):
        self._fixture_config_live_minimal()
        self._fixture_config_local_daemons(
            corosync_enabled=True, corosync_active=True,
            pacemaker_enabled=True, pacemaker_active=True,
            pacemaker_remote_enabled=True, pacemaker_remote_active=True,
            pcsd_enabled=True, pcsd_active=True,
            sbd_enabled=True, sbd_active=True,
        )
        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            dedent("""\
                Cluster name: test99
                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pacemaker_remote: active/enabled
                  pcsd: active/enabled
                  sbd: active/enabled"""
            )
        )

    def test_daemon_status_all_off(self):
        self._fixture_config_live_minimal()
        self._fixture_config_local_daemons(
            corosync_enabled=False, corosync_active=False,
            pacemaker_enabled=False, pacemaker_active=False,
            pacemaker_remote_enabled=False, pacemaker_remote_active=False,
            pcsd_enabled=False, pcsd_active=False,
            sbd_enabled=False, sbd_active=False,
        )
        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            dedent("""\
                Cluster name: test99
                crm_mon cluster status

                Daemon Status:
                  corosync: inactive/disabled
                  pacemaker: inactive/disabled
                  pcsd: inactive/disabled"""
            )
        )



from typing import (
    Any,
    Mapping,
    Type,
    TypeVar,
)


class DataTransferObject:
    def to_dict(self) -> Mapping[str, Any]:
        raise NotImplementedError()

    @classmethod
    def from_dict(cls, payload: Mapping[str, Any]) -> "DataTransferObject":
        raise NotImplementedError()


class ImplementsToDto:
    def to_dto(self) -> DataTransferObject:
        raise NotImplementedError()


T = TypeVar("T")

class ImplementsFromDto:
    @classmethod
    def from_dto(cls: Type[T], dto_obj: DataTransferObject) -> T:
        raise NotImplementedError()



"""
This module defines madatory and optional cib sections. It provides function for
getting existing section from the cib (lxml) tree.
"""
from pcs.lib import reports
from pcs.lib.errors import LibraryError
from pcs.lib.xml_tools import get_sub_element


CONFIGURATION = "configuration"
CONSTRAINTS = "configuration/constraints"
CRM_CONFIG = "configuration/crm_config"
NODES = "configuration/nodes"
RESOURCES = "configuration/resources"

ACLS = "acls"
ALERTS = "alerts"
FENCING_TOPOLOGY = "fencing-topology"
OP_DEFAULTS = "op_defaults"
RSC_DEFAULTS = "rsc_defaults"

__MANDATORY_SECTIONS = [
    CONFIGURATION,
    CONSTRAINTS,
    CRM_CONFIG,
    NODES,
    RESOURCES,
]

__OPTIONAL_SECTIONS = [
    ACLS,
    ALERTS,
    FENCING_TOPOLOGY,
    OP_DEFAULTS,
    RSC_DEFAULTS,
]

def get(tree, section_name):
    """
    Return the element which represents section 'section_name' in the tree.

    If the section is mandatory and is not found in the tree this function
    raises.
    If the section is optional and is not found in the tree this function
    creates new section.

    lxml.etree.Element tree -- is tree in which the section is looked up
    string section_name -- name of desired section; it is strongly recommended
        to use constants defined in this module
    """
    if section_name in __MANDATORY_SECTIONS:
        section = tree.find(".//{0}".format(section_name))
        if section is not None:
            return section
        raise LibraryError(reports.cib_missing_mandatory_section(section_name))

    if section_name in __OPTIONAL_SECTIONS:
        return get_sub_element(get(tree, CONFIGURATION), section_name)

    raise AssertionError("Unknown cib section '{0}'".format(section_name))

def exists(tree, section_name):
    if section_name not in __MANDATORY_SECTIONS + __OPTIONAL_SECTIONS:
        raise AssertionError("Unknown cib section '{0}'".format(section_name))
    return tree.find(".//{0}".format(section_name)) is not None



from pcs import settings
from pcs.common import file_type_codes
from pcs.common.file import RawFileError
from pcs.common.reports import SimpleReportProcessor
from pcs.common.tools import format_environment_error
from pcs.lib import reports
from pcs.lib.communication.nodes import SendPcsdSslCertAndKey
from pcs.lib.communication.tools import run_and_raise
from pcs.lib.errors import LibraryError
from pcs.lib.node import get_existing_nodes_names

def synchronize_ssl_certificate(env, skip_offline=False):
    """
    Send the local pcsd SSL cert and key to all full nodes in the local cluster.

    Consider the pcs Web UI is accessed via an IP running as a resource in the
    cluster. When the IP is moved, the user's browser connects to the new node
    and we want it to get the same certificate to make the transition a
    seamless experience (otherwise the browser display a warning that the
    certificate has changed).
    Using pcsd Web UI on remote and guest nodes is not supported (pcs/pcsd
    depends on the corosanc.conf file being present on the local node) so we
    send the cert only to corossync (== full stack) nodes.
    """
    report_processor = SimpleReportProcessor(env.report_processor)
    target_factory = env.get_node_target_factory()
    cluster_nodes_names, report_list = get_existing_nodes_names(
        env.get_corosync_conf()
    )
    if not cluster_nodes_names:
        report_list.append(reports.corosync_config_no_nodes_defined())
    report_processor.report_list(report_list)

    try:
        with open(settings.pcsd_cert_location, "r") as file:
            ssl_cert = file.read()
    except EnvironmentError as e:
        report_processor.report(
            reports.file_io_error(
                file_type_codes.PCSD_SSL_CERT,
                RawFileError.ACTION_READ,
                format_environment_error(e),
                file_path=settings.pcsd_cert_location,
            )
        )
    try:
        with open(settings.pcsd_key_location, "r") as file:
            ssl_key = file.read()
    except EnvironmentError as e:
        report_processor.report(
            reports.file_io_error(
                file_type_codes.PCSD_SSL_KEY,
                RawFileError.ACTION_READ,
                format_environment_error(e),
                file_path=settings.pcsd_key_location,
            )
        )

    target_report_list, target_list = (
        target_factory.get_target_list_with_reports(
            cluster_nodes_names,
            skip_non_existing=skip_offline
        )
    )
    report_processor.report_list(target_report_list)

    if report_processor.has_errors:
        raise LibraryError()

    env.report_processor.process(
        reports.pcsd_ssl_cert_and_key_distribution_started(
            [target.label for target in target_list]
        )
    )

    com_cmd = SendPcsdSslCertAndKey(env.report_processor, ssl_cert, ssl_key)
    com_cmd.set_targets(target_list)
    run_and_raise(env.get_node_communicator(), com_cmd)



from unittest import mock, TestCase

from pcs_test.tools.misc import get_test_resource as rc

from pcs.cli.common import capabilities


@mock.patch("pcs.settings.pcsd_exec_location", rc(""))
class Capabilities(TestCase):
    def test_get_definition(self):
        self.assertEqual(
            capabilities.get_capabilities_definition(),
            [
                {
                    "id": "test.in-pcs",
                    "in-pcs": "1",
                    "in-pcsd": "0",
                    "description": "This capability is available in pcs.",
                },
                {
                    "id": "test.in-pcsd",
                    "in-pcs": "0",
                    "in-pcsd": "1",
                    "description": "This capability is available in pcsd.",
                },
                {
                    "id": "test.both",
                    "in-pcs": "1",
                    "in-pcsd": "1",
                    "description":
                        "This capability is available in both pcs and pcsd.",
                },
                {
                    "id": "test.empty-description",
                    "in-pcs": "1",
                    "in-pcsd": "1",
                    "description": "",
                },
                {
                    "id": "test.no-description",
                    "in-pcs": "1",
                    "in-pcsd": "1",
                    "description": "",
                },
            ]
        )

    def test_get_pcs(self):
        self.assertEqual(
            capabilities.get_pcs_capabilities(),
            [
                {
                    "id": "test.in-pcs",
                    "description": "This capability is available in pcs.",
                },
                {
                    "id": "test.both",
                    "description":
                        "This capability is available in both pcs and pcsd.",
                },
                {
                    "id": "test.empty-description",
                    "description": "",
                },
                {
                    "id": "test.no-description",
                    "description": "",
                },
            ]
        )



from tornado.web import RequestHandler

class EnhanceHeadersMixin:
    """
    EnhanceHeadersMixin allows to add security headers to GUI urls.
    """
    def set_strict_transport_security(self):
        # rhbz 1558063
        # The HTTP Strict-Transport-Security response header (often abbreviated
        # as HSTS)  lets a web site tell browsers that it should only be
        # accessed using HTTPS, instead of using HTTP.
        self.set_header("Strict-Transport-Security", "max-age=604800")

    def set_header_nosniff_content_type(self):
        # The X-Content-Type-Options response HTTP header is a marker used by
        # the server to indicate that the MIME types advertised in the
        # Content-Type headers should not be changed and be followed. This
        # allows to opt-out of MIME type sniffing, or, in other words, it is a
        # way to say that the webmasters knew what they were doing.
        self.set_header("X-Content-Type-Options", "nosniff")

    def enhance_headers(self):
        self.set_header_nosniff_content_type()

        # The X-Frame-Options HTTP response header can be used to indicate
        # whether or not a browser should be allowed to render a page in a
        # <frame>, <iframe> or <object> . Sites can use this to avoid
        # clickjacking attacks, by ensuring that their content is not embedded
        # into other sites.
        self.set_header("X-Frame-Options", "SAMEORIGIN")

        # The HTTP X-XSS-Protection response header is a feature of Internet
        # Explorer, Chrome and Safari that stops pages from loading when they
        # detect reflected cross-site scripting (XSS) attacks. Although these
        # protections are largely unnecessary in modern browsers when sites
        # implement a strong Content-Security-Policy that disables the use of
        # inline JavaScript ('unsafe-inline'), they can still provide
        # protections for users of older web browsers that don't yet support
        # CSP.
        self.set_header("X-Xss-Protection", "1; mode=block")

class BaseHandler(EnhanceHeadersMixin, RequestHandler):
    """
    BaseHandler adds for all urls Strict-Transport-Security.
    """
    def set_default_headers(self):
        self.set_strict_transport_security()

    def data_received(self, chunk):
        # abstract method `data_received` does need to be overriden. This
        # method should be implemented to handle streamed request data.
        # BUT we currently do not plan to use it SO:
        #pylint: disable=abstract-method
        pass



from functools import partial
from unittest import mock, TestCase
from lxml import etree

from pcs_test.tools import fixture
from pcs_test.tools.assertions import  assert_xml_equal
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import (
    get_test_resource as rc,
    create_setup_patch_mixin,
)
from pcs_test.tools.xml import etree_to_str

from pcs.common import report_codes
from pcs.common.tools import Version
from pcs.lib.env import LibraryEnvironment


def mock_tmpfile(filename):
    mock_file = mock.MagicMock()
    mock_file.name = rc(filename)
    return mock_file

SetupPatchMixin = create_setup_patch_mixin(
    partial(mock.patch.object, LibraryEnvironment)
)


class ManageCibAssertionMixin:
    def assert_raises_cib_error(self, callable_obj, message):
        with self.assertRaises(AssertionError) as context_manager:
            callable_obj()
            self.assertEqual(str(context_manager.exception), message)

    def assert_raises_cib_not_loaded(self, callable_obj):
        self.assert_raises_cib_error(
            callable_obj,
            "CIB has not been loaded"
        )

    def assert_raises_cib_already_loaded(self, callable_obj):
        self.assert_raises_cib_error(
            callable_obj,
            "CIB has already been loaded"
        )

    def assert_raises_cib_loaded_cannot_custom(self, callable_obj):
        self.assert_raises_cib_error(
            callable_obj,
            "CIB has been loaded, cannot push custom CIB"
        )


class IsCibLive(TestCase):
    def test_is_live_when_no_cib_data_specified(self):
        env_assist, _ = get_env_tools(test_case=self)
        self.assertTrue(env_assist.get_env().is_cib_live)

    def test_is_not_live_when_cib_data_specified(self):
        env_assist, config = get_env_tools(test_case=self)
        config.env.set_cib_data("<cib/>")
        self.assertFalse(env_assist.get_env().is_cib_live)


class UpgradeCib(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_get_and_push_cib_version_upgrade_needed(self):
        (self.config
            .runner.cib.load(name="load_cib_old", filename="cib-empty-2.6.xml")
            .runner.cib.upgrade()
            .runner.cib.load(filename="cib-empty-2.8.xml")
        )
        env = self.env_assist.get_env()
        env.get_cib(Version(2, 8, 0))

        self.env_assist.assert_reports(
            [fixture.info(report_codes.CIB_UPGRADE_SUCCESSFUL)]
        )

    def test_get_and_push_cib_version_upgrade_not_needed(self):
        self.config.runner.cib.load(filename="cib-empty-2.6.xml")
        env = self.env_assist.get_env()
        env.get_cib(Version(2, 5, 0))


class GetCib(TestCase, ManageCibAssertionMixin):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_raise_library_error_when_cibadmin_failed(self):
        stderr = "cibadmin: Connection to local file failed..."
        (self.config
            #Value of cib_data is unimportant here. This content is only put
            #into tempfile when the runner is not mocked. And content is then
            #loaded from tempfile by `cibadmin --local --query`. Runner is
            #mocked in tests so the value of cib_data is not in the fact used.
            .env.set_cib_data("whatever")
            .runner.cib.load(returncode=203, stderr=stderr)
        )

        self.env_assist.assert_raise_library_error(
            self.env_assist.get_env().get_cib,
            [
                fixture.error(report_codes.CIB_LOAD_ERROR, reason=stderr)
            ],
            expected_in_processor=False
        )

    def test_returns_cib_from_cib_data(self):
        cib_filename = "cib-empty.xml"
        (self.config
            #Value of cib_data is unimportant here. See details in sibling test.
            .env.set_cib_data("whatever")
            .runner.cib.load(filename=cib_filename)
        )
        with open(rc(cib_filename)) as cib_file:
            assert_xml_equal(
                etree_to_str(self.env_assist.get_env().get_cib()),
                cib_file.read()
            )

    def test_get_and_property(self):
        self.config.runner.cib.load()
        env = self.env_assist.get_env()
        self.assertEqual(env.get_cib(), env.cib)

    def test_property_without_get(self):
        env = self.env_assist.get_env()
        # need to use lambda because env.cib is a property
        self.assert_raises_cib_not_loaded(lambda: env.cib)

    def test_double_get(self):
        self.config.runner.cib.load()
        env = self.env_assist.get_env()
        env.get_cib()
        self.assert_raises_cib_already_loaded(env.get_cib)


class PushLoadedCib(TestCase, ManageCibAssertionMixin):
    # pylint: disable=too-many-public-methods
    wait_timeout = 10
    def setUp(self):
        tmpfile_patcher = mock.patch("pcs.lib.pacemaker.live.write_tmpfile")
        self.addCleanup(tmpfile_patcher.stop)
        self.mock_write_tmpfile = tmpfile_patcher.start()
        self.tmpfile_old = mock_tmpfile("old.cib")
        self.tmpfile_new = mock_tmpfile("new.cib")
        self.mock_write_tmpfile.side_effect = [
            self.tmpfile_old, self.tmpfile_new
        ]
        self.cib_can_diff = "cib-empty-2.0.xml"
        self.cib_cannot_diff = "cib-empty-1.2.xml"
        self.env_assist, self.config = get_env_tools(test_case=self)

    def config_load_and_push_diff(self):
        (self.config
            .runner.cib.load(filename=self.cib_can_diff)
            .runner.cib.diff(self.tmpfile_old.name, self.tmpfile_new.name)
            .runner.cib.push_diff()
        )

    def config_load_and_push(self):
        (self.config
            .runner.cib.load(filename=self.cib_cannot_diff)
            .runner.cib.push()
        )

    def push_reports(self, cib_old=None, cib_new=None):
        # No test changes the CIB between load and push. The point is to test
        # loading and pushing, not editing the CIB.
        loaded_cib = self.config.calls.get("runner.cib.load").stdout
        return [
            fixture.debug(
                report_codes.TMP_FILE_WRITE,
                file_path=self.tmpfile_old.name,
                content=(cib_old if cib_old is not None else loaded_cib)
            ),
            fixture.debug(
                report_codes.TMP_FILE_WRITE,
                file_path=self.tmpfile_new.name,
                content=(cib_new if cib_new is not None else loaded_cib).strip()
            ),
        ]

    @staticmethod
    def push_full_forced_reports(version):
        return [
            fixture.warn(
                report_codes.CIB_PUSH_FORCED_FULL_DUE_TO_CRM_FEATURE_SET,
                current_set=version,
                required_set="3.0.9"
            )
        ]

    def test_get_and_push(self):
        self.config_load_and_push_diff()
        env = self.env_assist.get_env()

        env.get_cib()
        env.push_cib()
        self.env_assist.assert_reports(self.push_reports())

    def test_get_and_push_cannot_diff(self):
        self.config_load_and_push()
        env = self.env_assist.get_env()

        env.get_cib()
        env.push_cib()
        self.env_assist.assert_reports(
            self.push_full_forced_reports("3.0.8")
        )

    def test_modified_cib_features_do_not_matter(self):
        self.config_load_and_push_diff()
        env = self.env_assist.get_env()

        cib = env.get_cib()
        cib.set("crm_feature_set", "3.0.8")
        env.push_cib()
        self.env_assist.assert_reports(self.push_reports(
            cib_new=self.config.calls.get("runner.cib.load").stdout.replace(
                "3.0.9",
                "3.0.8"
            )
        ))

    def test_push_no_features_goes_with_full(self):
        (self.config
            .runner.cib.load_content("<cib />", name="runner.cib.load_content")
            .runner.cib.push(load_key="runner.cib.load_content")
        )
        env = self.env_assist.get_env()

        env.get_cib()
        env.push_cib()
        self.env_assist.assert_reports(
            self.push_full_forced_reports("0.0.0")
        )

    def test_can_get_after_push(self):
        self.config_load_and_push_diff()
        self.config.runner.cib.load(
            name="load_cib_2",
            filename=self.cib_can_diff
        )
        env = self.env_assist.get_env()

        env.get_cib()
        env.push_cib()
        # need to use lambda because env.cib is a property
        self.assert_raises_cib_not_loaded(lambda: env.cib)
        env.get_cib()
        self.env_assist.assert_reports(self.push_reports())

    def test_can_get_after_push_cannot_diff(self):
        self.config_load_and_push()
        self.config.runner.cib.load(
            name="load_cib_2",
            filename=self.cib_cannot_diff
         )
        env = self.env_assist.get_env()

        env.get_cib()
        env.push_cib()
        # need to use lambda because env.cib is a property
        self.assert_raises_cib_not_loaded(lambda: env.cib)
        env.get_cib()
        self.env_assist.assert_reports(
            self.push_full_forced_reports("3.0.8")
        )

    def test_not_loaded(self):
        env = self.env_assist.get_env()
        self.assert_raises_cib_not_loaded(env.push_cib)

    def test_tmpfile_fails(self):
        self.config.runner.cib.load(filename=self.cib_can_diff)
        self.mock_write_tmpfile.side_effect = EnvironmentError("test error")
        env = self.env_assist.get_env()

        env.get_cib()
        self.env_assist.assert_raise_library_error(
            env.push_cib,
            [
                fixture.error(
                    report_codes.CIB_SAVE_TMP_ERROR,
                    reason="test error",
                )
            ],
            expected_in_processor=False
        )

    def test_diff_is_empty(self):
        (self.config
            .runner.cib.load(filename=self.cib_can_diff)
            .runner.cib.diff(
                self.tmpfile_old.name,
                self.tmpfile_new.name,
                stdout="",
                stderr="",
                returncode=1
            )
        )
        env = self.env_assist.get_env()
        env.get_cib()
        env.push_cib()
        self.env_assist.assert_reports(self.push_reports())

    def test_diff_fails(self):
        (self.config
            .runner.cib.load(filename=self.cib_can_diff)
            .runner.cib.diff(
                self.tmpfile_old.name,
                self.tmpfile_new.name,
                stderr="invalid cib",
                returncode=65
            )
        )
        env = self.env_assist.get_env()
        env.get_cib()
        self.env_assist.assert_raise_library_error(
            env.push_cib,
            [
                fixture.error(
                    report_codes.CIB_DIFF_ERROR,
                    reason="invalid cib",
                )
            ],
            expected_in_processor=False
        )
        self.env_assist.assert_reports(self.push_reports())

    def test_push_diff_fails(self):
        (self.config
            .runner.cib.load(filename=self.cib_can_diff)
            .runner.cib.diff(self.tmpfile_old.name, self.tmpfile_new.name)
            .runner.cib.push_diff(stderr="invalid cib", returncode=1)
        )
        env = self.env_assist.get_env()
        env.get_cib()
        self.env_assist.assert_raise_library_error(
            env.push_cib,
            [
                fixture.error(
                    report_codes.CIB_PUSH_ERROR,
                    reason="invalid cib",
                )
            ],
            expected_in_processor=False
        )
        self.env_assist.assert_reports(self.push_reports())

    def test_push_fails(self):
        (self.config
            .runner.cib.load(filename=self.cib_cannot_diff)
            .runner.cib.push(stderr="invalid cib", returncode=1)
        )
        env = self.env_assist.get_env()
        env.get_cib()
        self.env_assist.assert_raise_library_error(
            env.push_cib,
            [
                fixture.error(
                    report_codes.CIB_PUSH_ERROR,
                    reason="invalid cib",
                )
            ],
            expected_in_processor=False
        )
        self.env_assist.assert_reports(
            self.push_full_forced_reports("3.0.8")
        )

    def test_wait(self):
        (self.config
            .runner.cib.load(filename=self.cib_can_diff)
            .runner.pcmk.can_wait()
            .runner.cib.diff(self.tmpfile_old.name, self.tmpfile_new.name)
            .runner.cib.push_diff()
            .runner.pcmk.wait(timeout=self.wait_timeout)
        )
        env = self.env_assist.get_env()

        env.get_cib()
        env.push_cib(wait=self.wait_timeout)
        self.env_assist.assert_reports(self.push_reports())

    def test_wait_cannot_diff(self):
        (self.config
            .runner.cib.load(filename=self.cib_cannot_diff)
            .runner.pcmk.can_wait()
            .runner.cib.push()
            .runner.pcmk.wait(timeout=self.wait_timeout)
        )
        env = self.env_assist.get_env()

        env.get_cib()
        env.push_cib(wait=self.wait_timeout)
        self.env_assist.assert_reports(
            self.push_full_forced_reports("3.0.8")
        )

    def test_wait_not_supported(self):
        (self.config
            .runner.cib.load(filename=self.cib_can_diff)
            .runner.pcmk.can_wait(stdout="cannot wait")
        )
        env = self.env_assist.get_env()

        env.get_cib()
        self.env_assist.assert_raise_library_error(
            lambda: env.push_cib(wait=self.wait_timeout),
            [
                fixture.error(report_codes.WAIT_FOR_IDLE_NOT_SUPPORTED),
            ],
            expected_in_processor=False
        )

    def test_wait_raises_on_invalid_value(self):
        (self.config
            .runner.cib.load(filename=self.cib_can_diff)
            .runner.pcmk.can_wait()
        )
        env = self.env_assist.get_env()

        env.get_cib()
        self.env_assist.assert_raise_library_error(
            lambda: env.push_cib(wait="abc"),
            [
                fixture.error(
                    report_codes.INVALID_TIMEOUT_VALUE,
                    timeout="abc"
                ),
            ],
            expected_in_processor=False
        )


class PushCustomCib(TestCase, ManageCibAssertionMixin):
    custom_cib = "<custom_cib />"
    wait_timeout = 10
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_push_without_get(self):
        self.config.runner.cib.push_independent(self.custom_cib)
        self.env_assist.get_env().push_cib(etree.XML(self.custom_cib))

    def test_push_after_get(self):
        self.config.runner.cib.load()
        env = self.env_assist.get_env()

        env.get_cib()
        self.assert_raises_cib_loaded_cannot_custom(
            partial(env.push_cib, etree.XML(self.custom_cib))
        )

    def test_wait(self):
        (self.config
            .runner.pcmk.can_wait()
            .runner.cib.push_independent(self.custom_cib)
            .runner.pcmk.wait(timeout=self.wait_timeout)
        )
        env = self.env_assist.get_env()

        env.push_cib(
            etree.XML(self.custom_cib),
            wait=self.wait_timeout
        )

    def test_wait_not_supported(self):
        self.config.runner.pcmk.can_wait(stdout="cannot wait")
        env = self.env_assist.get_env()

        self.env_assist.assert_raise_library_error(
            lambda: env.push_cib(
                etree.XML(self.custom_cib),
                wait=self.wait_timeout
            ),
            [
                fixture.error(report_codes.WAIT_FOR_IDLE_NOT_SUPPORTED),
            ],
            expected_in_processor=False
        )

    def test_wait_raises_on_invalid_value(self):
        self.config.runner.pcmk.can_wait()
        env = self.env_assist.get_env()

        self.env_assist.assert_raise_library_error(
            lambda: env.push_cib(etree.XML(self.custom_cib), wait="abc"),
            [
                fixture.error(
                    report_codes.INVALID_TIMEOUT_VALUE,
                    timeout="abc"
                ),
            ],
            expected_in_processor=False
        )


class PushCibMockedWithWait(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_wait_not_suported_for_mocked_cib(self):
        (self.config
            .env.set_cib_data("<cib/>")
            .runner.cib.load()
        )

        env = self.env_assist.get_env()
        env.get_cib()
        self.env_assist.assert_raise_library_error(
            lambda: env.push_cib(wait=10),
            [
                fixture.error(report_codes.WAIT_FOR_IDLE_NOT_LIVE_CLUSTER),
            ],
            expected_in_processor=False
        )



# pylint: disable=import-error
from pyagentx import Updater
from pcs.snmp.agentx.types import Oid


class AgentxUpdaterBase(Updater):
    """
    Base class for SNMP angent updaters. It provides methods for comfortable
    setting of values provided by the agent.
    """

    # this has to be set by the descendants
    _oid_tree: Oid

    @property
    def oid_tree(self):
        return self._oid_tree

    def _set_val(self, data_type, oid, value):
        self._data[oid] = {'name': oid, 'type': data_type, 'value': value}

    def _set_value_list(self, data_type, oid, value):
        if not isinstance(value, list):
            value = [value]
        for index, val in enumerate(value):
            self._set_val(
                data_type, "{oid}.{index}".format(oid=oid, index=index), val
            )

    def set_typed_value(self, oid, value):
        """
        oid string -- oid in the number form
        value BaseType -- BaseType object filled with value to set. Value can be
          either primitive value or list of primitive values.
        """
        self._set_value_list(value.data_type, oid, value.value)

    def set_value(self, str_oid, value):
        """
        str_oid string -- string form of oid. Raw (number form) oid will be
          figured out based on oid_tree tree.
        value primitive value or list of primitive values -- value to be set on
          specified str_oid
        """
        oid, oid_cls = _str_oid_to_oid(self.oid_tree, str_oid)
        self.set_typed_value(oid, oid_cls.data_type(value))

    def set_table(self, oid, table):
        """
        oid string -- number form of oid
        table list of list of BaseType -- members of outer list represent rows
          of table and members of inner list are columns.
        """
        for row in table:
            if not row:
                continue
            row_id = _str_to_oid(str(row[0].value))
            for index, col in enumerate(row[1:], start=2):
                value_oid = "{base_oid}.{index}.{row_id}".format(
                    base_oid=oid, index=index, row_id=row_id
                )
                self._set_val(col.data_type, value_oid, col.value)


def _find_oid_in_sub_tree(sub_tree, section_name):
    if sub_tree.member_list is None:
        return None
    for oid in sub_tree.member_list:
        if oid.str_oid == section_name:
            return oid
    return None


def _str_oid_to_oid(sub_tree, str_oid):
    sections = str_oid.split(".")
    oid_list = []
    for section in sections:
        sub_tree = _find_oid_in_sub_tree(sub_tree, section)
        if sub_tree is None:
            raise AssertionError(
                "oid section '{0}' ({1}) not found in section '{2}'".format(
                    section, str_oid, oid_list[-1] if oid_list else "."
                )
            )
        oid_list.append(str(sub_tree.oid))
        if sub_tree.data_type:
            oid = ".".join(oid_list)
            return (oid, sub_tree)
    return None


def _str_to_oid(data):
    length = len(data)
    oid_int = [str(ord(i)) for i in data]
    return str(length) + '.' + '.'.join(oid_int)



import logging
import os
from unittest import mock
from lxml import etree

from pcs_test.tools.custom_mock import MockLibraryReportProcessor
from pcs_test.tools.xml import etree_to_str

from pcs import settings
from pcs.lib.external import CommandRunner

# pylint: disable=line-too-long

def wrap_element_by_master(cib_file, resource_id, master_id=None):
    cib_tree = etree.parse(cib_file, etree.XMLParser(huge_tree=True)).getroot()
    element = cib_tree.find(f'.//*[@id="{resource_id}"]')
    final_master_id = (
        master_id if master_id is not None else f"{resource_id}-master"
    )
    master_element = _xml_to_element(f"""
        <master id="{final_master_id}">
        </master>
    """)
    element.getparent().append(master_element)
    master_element.append(element)
    final_xml = etree_to_str(cib_tree)

    environ = dict(os.environ)
    environ["CIB_file"] = cib_file
    runner = CommandRunner(
        mock.MagicMock(logging.Logger),
        MockLibraryReportProcessor(),
        environ
    )
    stdout, stderr, retval = runner.run([
        os.path.join(settings.pacemaker_binaries, "cibadmin"),
        "--replace", "--scope", "resources", "--xml-pipe",
    ], stdin_string=final_xml)
    assert retval == 0, (
        "Error running wrap_element_by_master:\n" + stderr + "\n" + stdout
    )

def fixture_master_xml(name, all_ops=True, meta_dict=None):
    default_ops = f"""
            <op id="{name}-notify-interval-0s" interval="0s" name="notify"
                timeout="5"
            />
            <op id="{name}-start-interval-0s" interval="0s" name="start"
                timeout="20"
            />
            <op id="{name}-stop-interval-0s" interval="0s" name="stop"
                timeout="20"
            />
    """
    meta_xml = ''
    if meta_dict:
        meta_lines = (
            [f'<meta_attributes id="{name}-master-meta_attributes">']
            +
            [
                f'<nvpair id="{name}-master-meta_attributes-{key}" name="{key}" value="{val}"/>'
                for key, val in meta_dict.items()
            ]
            +
            ['</meta_attributes>']
        )
        meta_xml = "\n".join(meta_lines)
    master = f"""
      <master id="{name}-master">
        <primitive class="ocf" id="{name}" provider="pacemaker" type="Stateful">
          <operations>
            <op id="{name}-monitor-interval-10" interval="10" name="monitor"
                role="Master" timeout="20"
            />
            <op id="{name}-monitor-interval-11" interval="11" name="monitor"
                role="Slave" timeout="20"
            />
    """
    if all_ops:
        master += default_ops
    master += f"""
          </operations>
        </primitive>
        {meta_xml}
      </master>
    """
    return master

def fixture_to_cib(cib_file, xml):
    environ = dict(os.environ)
    environ["CIB_file"] = cib_file
    runner = CommandRunner(
        mock.MagicMock(logging.Logger),
        MockLibraryReportProcessor(),
        environ
    )
    stdout, stderr, retval = runner.run([
        os.path.join(settings.pacemaker_binaries, "cibadmin"),
        "--create", "--scope", "resources", "--xml-text", xml
    ])
    assert retval == 0, (
        "Error running fixture_to_cib:\n" + stderr + "\n" + stdout
    )

def _replace(element_to_replace, new_element):
    parent = element_to_replace.getparent()
    for child in parent:
        if element_to_replace == child:
            index = list(parent).index(child)
            parent.remove(child)
            parent.insert(index, new_element)
            return

def _xml_to_element(xml):
    try:
        new_element = etree.fromstring(xml)
    except etree.XMLSyntaxError:
        raise AssertionError(
            "Cannot put to the cib a non-xml fragment:\n'{0}'"
            .format(xml)
        )
    return new_element

def _find_all_in(cib_tree, element_xpath):
    element_list = cib_tree.xpath(element_xpath)
    if not element_list:
        raise AssertionError(
            "Cannot find '{0}' in given cib:\n{1}".format(
                element_xpath,
                etree_to_str(cib_tree)
            )
        )
    return element_list

def _find_in(cib_tree, element_xpath):
    element_list = _find_all_in(cib_tree, element_xpath)
    if len(element_list) > 1:
        raise AssertionError(
            "Found more than one '{0}' in given cib:\n{1}".format(
                element_xpath,
                etree_to_str(cib_tree)
            )
        )
    return element_list[0]

def remove(element_xpath):
    def _remove(cib_tree):
        xpath_list = (
            [element_xpath] if isinstance(element_xpath, str) else element_xpath
        )
        for xpath in xpath_list:
            for element_to_remove in _find_all_in(cib_tree, xpath):
                element_to_remove.getparent().remove(element_to_remove)
    return _remove

def put_or_replace(parent_xpath, new_content):
    #This tranformation makes sense in "configuration" section only. In this
    #section there are sub-tags (optional or mandatory) that can occure max 1x.
    #
    #In other sections it is possible to have more occurences of sub-tags. For
    #such cases it is better to use `replace_all` - the difference is that in
    #`replace_all` the element to be replaced is specified by full xpath
    #whilst in `put_or_replace` the xpath to the parent element is specified.
    def replace_optional(cib_tree):
        element = _xml_to_element(new_content)
        parent = _find_in(cib_tree, parent_xpath)
        current_elements = parent.findall(element.tag)

        if len(current_elements) > 1:
            raise _cannot_multireplace(element.tag, parent_xpath, cib_tree)

        if current_elements:
            _replace(current_elements[0], element)
        else:
            parent.append(element)

    return replace_optional

def replace_all(replacements):
    """
    Return a function that replace more elements (defined by replacement_dict)
    in the cib_tree with new_content.

    dict replacemens -- contains more replacements:
        key is xpath - its destination must be one element: replacement is
        applied only on the first occurence
        value is new content -contains a content that have to be placed instead
        of an element found by element_xpath
    """
    def replace(cib_tree):
        for xpath, new_content in replacements.items():
            _replace(_find_in(cib_tree, xpath), _xml_to_element(new_content))
    return replace

def append_all(append_map):
    """
    Return a function that appends more elements after specified (xpath) element
    dict append_map -- a key is an xpath pointing to a target element (for
        appending), value is appended content
    """
    def append(cib_tree):
        for xpath, new_content in append_map.items():
            _find_in(cib_tree, xpath).append(_xml_to_element(new_content))
    return append


#Possible modifier shortcuts are defined here.
#Keep in mind that every key will be named parameter in config function
#(see modifier_shortcuts param in some of pcs_test.tools.command_env.config_*
#modules)
#
#DO NOT USE CONFLICTING KEYS HERE!
#1) args of pcs_test.tools.command_env.calls#CallListBuilder.place:
#  name, before, instead
#2) args of pcs_test.tools.command_env.mock_runner#Call.__init__
#  command, stdout, stderr, returncode, check_stdin
#3) special args of pcs_test.tools.command_env.config_*
#  modifiers, filename, load_key, wait, exception
#It would be not aplied. Not even mention that the majority of these names do
#not make sense for a cib modifying ;)
MODIFIER_GENERATORS = {
    "remove": remove,
    "replace": replace_all,
    "append": append_all,
    "resources": lambda xml: replace_all({"./configuration/resources": xml}),
    "constraints": lambda xml: replace_all(
        {"./configuration/constraints": xml}
    ),
    "optional_in_conf": lambda xml: put_or_replace("./configuration", xml),
    #common modifier `put_or_replace` makes not sense - see explanation inside
    #this function - all occurences should be satisfied by `optional_in_conf`
}

def create_modifiers(**modifier_shortcuts):
    """
    Return list of modifiers: list of functions that transform cib

    dict modifier_shortcuts -- a new modifier is generated from each modifier
        shortcut.
        As key there can be keys of MODIFIER_GENERATORS.
        Value is passed into appropriate generator from MODIFIER_GENERATORS.

    """
    unknown_shortcuts = (
        set(modifier_shortcuts.keys()) - set(MODIFIER_GENERATORS.keys())
    )
    if unknown_shortcuts:
        raise AssertionError(
            "Unknown modifier shortcuts '{0}', available are: '{1}'".format(
                "', '".join(list(unknown_shortcuts)),
                "', '".join(MODIFIER_GENERATORS.keys()),
            )
        )

    return [
        MODIFIER_GENERATORS[name](param)
        for name, param in modifier_shortcuts.items()
    ]

def modify_cib(cib_xml, modifiers=None, **modifier_shortcuts):
    """
    Apply modifiers to cib_xml and return the result cib_xml

    string cib_xml -- initial cib
    list of callable modifiers -- each takes cib (etree.Element)
    dict modifier_shortcuts -- a new modifier is generated from each modifier
        shortcut.
        As key there can be keys of MODIFIER_GENERATORS.
        Value is passed into appropriate generator from MODIFIER_GENERATORS.
    """
    modifiers = modifiers if modifiers else []
    all_modifiers = modifiers + create_modifiers(**modifier_shortcuts)

    if not all_modifiers:
        return cib_xml

    cib_tree = etree.fromstring(cib_xml)
    for modify in all_modifiers:
        modify(cib_tree)

    return etree_to_str(cib_tree)

def _cannot_multireplace(tag, parent_xpath, cib_tree):
    return AssertionError(
        (
            "Cannot replace '{element}' in '{parent}' because '{parent}'"
            " contains more than one '{element}' in given cib:\n{cib}"
        ).format(element=tag, parent=parent_xpath, cib=etree_to_str(cib_tree))
    )



from pcs.lib.pacemaker.values import is_true

def constraint_plain(constraint_info, with_id=False):
    """
    dict constraint_info see constraint in pcs/lib/exchange_formats.md
    bool with_id have to show id with options_dict
    """
    options = constraint_info["options"]
    oc_resource1 = options.get("first", "")
    oc_resource2 = options.get("then", "")
    first_action = options.get("first-action", "")
    then_action = options.get("then-action", "")
    oc_id = options.get("id", "")
    oc_score = options.get("score", "")
    oc_kind = options.get("kind", "")
    oc_sym = ""
    oc_id_out = ""
    oc_options = ""
    if (
        "symmetrical" in options
        and
        not is_true(options.get("symmetrical", "false"))
    ):
        oc_sym = "(non-symmetrical)"
    if oc_kind != "":
        score_text = "(kind:" + oc_kind + ")"
    elif oc_kind == "" and oc_score == "":
        score_text = "(kind:Mandatory)"
    else:
        score_text = "(score:" + oc_score + ")"
    if with_id:
        oc_id_out = "(id:"+oc_id+")"
    already_processed_options = (
        "first", "then", "first-action", "then-action", "id", "score", "kind",
        "symmetrical"
    )
    oc_options = " ".join([
        "{0}={1}".format(name, value)
        for name, value in options.items()
        if name not in already_processed_options
    ])
    if oc_options:
        oc_options = "(Options: " + oc_options + ")"
    return " ".join([arg for arg in [
        first_action, oc_resource1, "then", then_action, oc_resource2,
        score_text, oc_sym, oc_options, oc_id_out
    ] if arg])



from unittest import TestCase

class CheckCorosyncOffline(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.test_env.PushCorosyncConfLiveNoQdeviceTest
    """

class DistributeCorosyncConf(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.test_env.PushCorosyncConfLiveNoQdeviceTest
    """

class ReloadCorosyncConf(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.cluster.test_add_nodes.{
            AddNodesSuccessMinimal
            AddNodeFull
        }   FailureReloadCorosyncConf
        pcs_test.tier0.lib.commands.cluster.test_remove_nodes
            .FailureCorosyncReload
        pcs_test.tier0.lib.test_env
    """



from unittest import TestCase

from pcs.lib import errors


class ReportListAnalyzerSelectSeverities(TestCase):
    def setUp(self):
        self.severities = [
            errors.ReportItemSeverity.WARNING,
            errors.ReportItemSeverity.INFO,
            errors.ReportItemSeverity.DEBUG,
        ]

    def assert_select_reports(self, all_reports, expected_errors):
        self.assertEqual(
            expected_errors,
            errors.ReportListAnalyzer(all_reports)
                .reports_with_severities(self.severities)
        )

    def test_returns_empty_on_no_reports(self):
        self.assert_select_reports([], [])

    def test_returns_empty_on_reports_with_other_severities(self):
        self.assert_select_reports([errors.ReportItem.error("ERR")], [])

    def test_returns_selection_of_desired_severities(self):
        err = errors.ReportItem.error("ERR")
        warn = errors.ReportItem.warning("WARN")
        info = errors.ReportItem.info("INFO")
        debug = errors.ReportItem.debug("DEBUG")
        self.assert_select_reports(
            [
                err,
                warn,
                info,
                debug,
            ],
            [
                warn,
                info,
                debug,
            ]
        )

class ReportListAnalyzerErrorList(TestCase):
    def assert_select_reports(self, all_reports, expected_errors):
        self.assertEqual(
            expected_errors,
            errors.ReportListAnalyzer(all_reports).error_list
        )

    def test_returns_empty_on_no_reports(self):
        self.assert_select_reports([], [])

    def test_returns_empty_on_no_errors(self):
        self.assert_select_reports([errors.ReportItem.warning("WARN")], [])

    def test_returns_only_errors_on_mixed_content(self):
        err = errors.ReportItem.error("ERR")
        self.assert_select_reports(
            [errors.ReportItem.warning("WARN"), err],
            [err]
        )



from unittest import TestCase


class CheckReachability(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.command.status
    """

class GetOnlineTargets(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.sbd.test_enable_sbd
    """

class SendPcsdSslCertAndKey(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.cluster.test_setup
        pcs_test.tier0.lib.commands.test_pcsd
    """

class RemoveNodesFromCib(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.cluster.test_remove_nodes.{
            RemoveNodesFailureFromCib
            RemoveNodesSuccessMinimal
        }
    """



from unittest import TestCase

class GetCaCert(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.AddDeviceNetTest
    """

class ClientSetup(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.AddDeviceNetTest
    """

class SignCertificate(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.AddDeviceNetTest
    """

class ClientImportCertificateAndKey(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.AddDeviceNetTest
    """

class ClientDestroy(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.RemoveDeviceNetTest
    """



import os
import ssl

from OpenSSL import crypto, SSL

from pcs.common.ssl import (
    dump_cert,
    dump_key,
    generate_cert,
    generate_key,
)

def check_cert_key(cert_path, key_path):
    errors = []
    def load(load_ssl_file, label, path):
        try:
            with open(path) as ssl_file:
                return load_ssl_file(crypto.FILETYPE_PEM, ssl_file.read())
        except EnvironmentError as e:
            errors.append(f"Unable to read SSL {label} '{path}': '{e}'")
        except crypto.Error as e:
            msg = ""
            if e.args and e.args[0] and e.args[0][0]:
                msg = f": '{':'.join(e.args[0][0])}'"
            errors.append(f"Invalid SSL {label} '{path}'{msg}")

    cert = load(crypto.load_certificate, "certificate", cert_path)
    key = load(crypto.load_privatekey, "key", key_path)

    if errors:
        return errors

    try:
        context = SSL.Context(SSL.TLSv1_METHOD)
        context.use_privatekey(key)
        context.use_certificate(cert)
    except SSL.Error as e:
        errors.append(f"Unable to load SSL certificate and/or key: {e}")
        # If we cannot load the files, do not confuse users with other error
        # messages.
        return errors
    try:
        context.check_privatekey()
    except (crypto.Error, SSL.Error) as e:
        errors.append(f"SSL certificate does not match the key: {e}")

    return errors

def open_ssl_file_to_rewrite(path):
    return os.fdopen(os.open(path, os.O_CREAT|os.O_WRONLY, 0o600), "wb")

def regenerate_cert_key(server_name, cert_path, key_path, key_length=None):
    key = generate_key(key_length) if key_length else generate_key()
    with open_ssl_file_to_rewrite(cert_path) as cert_file:
        cert_file.write(dump_cert(generate_cert(key, server_name)))
    with open_ssl_file_to_rewrite(key_path) as key_file:
        key_file.write(dump_key(key))

class CertKeyPair:
    def __init__(self, cert_location, key_location):
        self.__cert_location = cert_location
        self.__key_location = key_location

    @property
    def cert_location(self):
        return self.__cert_location

    @property
    def key_location(self):
        return self.__key_location

    def exists(self):
        return (
            os.path.isfile(self.cert_location)
            and
            os.path.isfile(self.key_location)
        )

    def check(self):
        return check_cert_key(self.cert_location, self.key_location)

    def regenerate(self, server_name, key_length=None):
        regenerate_cert_key(
            server_name,
            self.cert_location,
            self.key_location,
            key_length
        )

class SSLCertKeyException(Exception):
    pass

class PcsdSSL:
    def __init__(
        self, server_name, cert_location, key_location, ssl_options, ssl_ciphers
    ):
        self.__server_name = server_name
        self.__ssl_options = ssl_options
        self.__ssl_ciphers = ssl_ciphers
        self.__ck_pair = CertKeyPair(cert_location, key_location)

    def create_context(self) -> ssl.SSLContext:
        # pylint: disable=no-member
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        ssl_context.set_ciphers(self.__ssl_ciphers)
        ssl_context.options = self.__ssl_options
        ssl_context.load_cert_chain(
            self.__ck_pair.cert_location,
            self.__ck_pair.key_location
        )
        return ssl_context

    def guarantee_valid_certs(self):
        if not self.__ck_pair.exists():
            self.__ck_pair.regenerate(self.__server_name)
            return

        error_list = self.__ck_pair.check()
        if error_list:
            raise SSLCertKeyException(*error_list)



from unittest import TestCase

from pcs.cli.common import middleware


class MiddlewareBuildTest(TestCase):
    def test_run_middleware_correctly_chained(self):
        log = []
        def command(lib, argv, modifiers):
            log.append('command: {0}, {1}, {2}'.format(lib, argv, modifiers))

        def mdw1(_next, lib, argv, modifiers):
            log.append(
                'mdw1 start: {0}, {1}, {2}'.format(lib, argv, modifiers)
            )
            _next(lib, argv, modifiers)
            log.append('mdw1 done')

        def mdw2(_next, lib, argv, modifiers):
            log.append(
                'mdw2 start: {0}, {1}, {2}'.format(lib, argv, modifiers)
            )
            _next(lib, argv, modifiers)
            log.append('mdw2 done')

        run_with_middleware = middleware.build(mdw1, mdw2)
        run_with_middleware(command, "1", "2", "3")
        self.assertEqual(log, [
            'mdw1 start: 1, 2, 3',
            'mdw2 start: 1, 2, 3',
            'command: 1, 2, 3',
            'mdw2 done',
            'mdw1 done',
        ])



def constraint_plain(constraint_type, constraint_info, with_id=False):
    return "{0} {1}".format(
        constraint_type,
        ' '.join(prepare_options(constraint_info['options'], with_id))
    )

def resource_sets(set_list, with_id=True):
    """
    list of dict set_list see resource set
        in pcs/lib/exchange_formats.md
    """
    report = []
    for resource_set in set_list:
        report.extend(
            ["set"] + resource_set["ids"] + options(resource_set["options"])
        )
        if with_id:
            report.append(id_from_options(resource_set["options"]))

    return report

def options(options_dict):
    return [
        key+"="+value
        for key, value in sorted(options_dict.items())
        if key != "id"
    ]

def id_from_options(options_dict):
    return "(id:"+options_dict.get("id", "")+")"

def constraint_with_sets(constraint_info, with_id=True):
    """
    dict constraint_info  see constraint in pcs/lib/exchange_formats.md
    bool with_id have to show id with options_dict
    """
    options_dict = options(constraint_info["options"])
    return " ".join(
        resource_sets(constraint_info["resource_sets"], with_id)
        +
        (["setoptions"] + options_dict if options_dict else [])
        +
        ([id_from_options(constraint_info["options"])] if with_id else [])
    )

def prepare_options(options_dict, with_id=True):
    return (
        options(options_dict)
        +
        ([id_from_options(options_dict)] if with_id else [])
    )



import base64
import json

from pcs.common import report_codes
from pcs.common.node_communicator import RequestData
from pcs.lib import reports
from pcs.lib.booth import reports as reports_booth
from pcs.lib.communication.tools import (
    AllAtOnceStrategyMixin,
    AllSameDataMixin,
    RunRemotelyBase,
    SkipOfflineMixin,
    SimpleResponseProcessingMixin,
)
from pcs.lib.errors import ReportItemSeverity


class BoothSendConfig(
    SimpleResponseProcessingMixin, SkipOfflineMixin, AllSameDataMixin,
    AllAtOnceStrategyMixin, RunRemotelyBase,
):
    # TODO adapt to new file transfer framework once it is written
    def __init__(
        self, report_processor, booth_name, config_data,
        authfile=None, authfile_data=None, skip_offline_targets=False
    ):
        super(BoothSendConfig, self).__init__(report_processor)
        self._set_skip_offline(skip_offline_targets)
        self._booth_name = booth_name
        self._config_data = config_data.decode("utf-8")
        self._authfile = authfile
        self._authfile_data = authfile_data

    def _get_request_data(self):
        data = {
            "config": {
                "name": "{0}.conf".format(self._booth_name),
                "data": self._config_data
            }
        }
        if self._authfile is not None and self._authfile_data is not None:
            data["authfile"] = {
                "name": self._authfile,
                "data": base64.b64encode(self._authfile_data).decode("utf-8")
            }
        return RequestData(
            "remote/booth_set_config", [("data_json", json.dumps(data))]
        )

    def _get_success_report(self, node_label):
        return reports_booth.booth_config_accepted_by_node(
            node_label, [self._booth_name]
        )

    def before(self):
        self._report(reports_booth.booth_config_distribution_started())


class ProcessJsonDataMixin:
    __data = None

    @property
    def _data(self):
        if self.__data is None:
            self.__data = []
        return self.__data

    def _process_response(self, response):
        report = self._get_response_report(response)
        if report is not None:
            self._report(report)
            return
        target = response.request.target
        try:
            self._data.append((target, json.loads(response.data)))
        except ValueError:
            self._report(reports.invalid_response_format(target.label))

    def on_complete(self):
        return self._data


class BoothGetConfig(
    ProcessJsonDataMixin, AllSameDataMixin, AllAtOnceStrategyMixin,
    RunRemotelyBase,
):
    def __init__(self, report_processor, booth_name):
        super(BoothGetConfig, self).__init__(report_processor)
        self._booth_name = booth_name

    def _get_request_data(self):
        return RequestData(
            "remote/booth_get_config", [("name", self._booth_name)]
        )


class BoothSaveFiles(
    ProcessJsonDataMixin, AllSameDataMixin, AllAtOnceStrategyMixin,
    RunRemotelyBase,
):
    def __init__(self, report_processor, file_list, rewrite_existing=True):
        super(BoothSaveFiles, self).__init__(report_processor)
        self._file_list = file_list
        self._rewrite_existing = rewrite_existing
        self._output_data = []

    def _get_request_data(self):
        data = [("data_json", json.dumps(self._file_list))]
        if self._rewrite_existing:
            data.append(("rewrite_existing", "1"))
        return RequestData("remote/booth_save_files", data)

    def _process_response(self, response):
        report = self._get_response_report(response)
        if report is not None:
            self._report(report)
            return
        target = response.request.target
        try:
            parsed_data = json.loads(response.data)
            self._report(
                reports_booth.booth_config_accepted_by_node(
                    target.label, list(parsed_data["saved"])
                )
            )
            for filename in list(parsed_data["existing"]):
                self._report(reports.file_already_exists(
                    "", # TODO specify file type; this will be overhauled to
                        # a generic file transport framework anyway
                    filename,
                    severity=(
                        ReportItemSeverity.WARNING if self._rewrite_existing
                        else ReportItemSeverity.ERROR
                    ),
                    forceable=(
                        None if self._rewrite_existing
                        else report_codes.FORCE_FILE_OVERWRITE
                    ),
                    node=target.label,
                ))
            for file, reason in dict(parsed_data["failed"]).items():
                self._report(reports_booth.booth_config_distribution_node_error(
                    target.label, reason, file
                ))
        except (KeyError, TypeError, ValueError):
            self._report(reports.invalid_response_format(target.label))



from typing import (
    Any,
    Iterable,
    List,
    Mapping,
    Sequence,
    Union,
)

from pcs.common.pacemaker.resource.relations import (
    RelationEntityDto,
    ResourceRelationDto,
    ResourceRelationType,
)
from pcs.cli.common.console_report import format_optional
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import InputModifiers
from pcs.cli.common.printable_tree import (
    tree_to_lines,
    PrintableTreeNode,
)


def show_resource_relations_cmd(
    lib: Any,
    argv: Sequence[str],
    modifiers: InputModifiers,
) -> None:
    """
    Options:
      * -f - CIB file
      * --full - show constraint ids and resource types
    """
    modifiers.ensure_only_supported("-f", "--full")
    if len(argv) != 1:
        raise CmdLineInputError()
    tree = ResourcePrintableNode.from_dto(
        ResourceRelationDto.from_dict(
            lib.resource.get_resource_relations_tree(argv[0])
        )
    )
    for line in tree_to_lines(tree, verbose=modifiers.get("--full")):
        print(line)


class ResourceRelationBase(PrintableTreeNode):
    def __init__(
        self,
        relation_entity: RelationEntityDto,
        members: Sequence["ResourceRelationBase"],
        is_leaf: bool,
    ):
        self._relation_entity = relation_entity
        self._members = members
        self._is_leaf = is_leaf

    @property
    def is_leaf(self) -> bool:
        return self._is_leaf

    @property
    def relation_entity(self) -> RelationEntityDto:
        return self._relation_entity

    @property
    def members(self) -> Sequence["ResourceRelationBase"]:
        return self._members

    @property
    def detail(self) -> Sequence[str]:
        raise NotImplementedError()

    def get_title(self, verbose: bool) -> str:
        raise NotImplementedError()



class ResourcePrintableNode(ResourceRelationBase):
    @classmethod
    def from_dto(
        cls, resource_dto: ResourceRelationDto
    ) -> "ResourcePrintableNode":
        def _relation_comparator(item: ResourceRelationBase) -> str:
            type_priorities = (
                ResourceRelationType.INNER_RESOURCES,
                ResourceRelationType.OUTER_RESOURCE,
                ResourceRelationType.ORDER,
                ResourceRelationType.ORDER_SET,
            )
            priority_map = {
                _type: value for value, _type in enumerate(type_priorities)
            }
            return "{_type}_{_id}".format(
                _type=priority_map.get(
                    # Hardcoded number 9 is intentional. If there is more than
                    # 10 items, it would be required to also prepend zeros for
                    # lower numbers. E.g: if there is 100 options, it should
                    # starts as 000, 001, ...
                    item.relation_entity.type, 9 # type: ignore
                ),
                _id=item.relation_entity.id
            )

        return cls(
            resource_dto.relation_entity,
            sorted(
                [
                    RelationPrintableNode.from_dto(member_dto)
                    for member_dto in resource_dto.members
                ],
                key=_relation_comparator,
            ),
            resource_dto.is_leaf
        )

    def get_title(self, verbose: bool) -> str:
        rsc_type = self._relation_entity.type
        metadata = self._relation_entity.metadata
        if rsc_type == "primitive":
            rsc_type = "{_class}{_provider}{_type}".format(
                _class=format_optional(metadata.get("class"), "{}:"),
                _provider=format_optional(metadata.get("provider"), "{}:"),
                _type=metadata.get("type"),
            )
        detail = f" (resource: {rsc_type})" if verbose else ""
        return f"{self._relation_entity.id}{detail}"

    @property
    def detail(self) -> Sequence[str]:
        return []


class RelationPrintableNode(ResourceRelationBase):
    @classmethod
    def from_dto(
        cls, relation_dto: ResourceRelationDto
    ) -> "RelationPrintableNode":
        return cls(
            relation_dto.relation_entity,
            sorted(
                [
                    ResourcePrintableNode.from_dto(member_dto)
                    for member_dto in relation_dto.members
                ],
                key=lambda item: item.relation_entity.id,
            ),
            relation_dto.is_leaf
        )

    def get_title(self, verbose: bool) -> str:
        rel_type_map: Mapping[Union[str, ResourceRelationType], str] = {
            ResourceRelationType.ORDER: "order",
            ResourceRelationType.ORDER_SET: "order set",
            ResourceRelationType.INNER_RESOURCES: "inner resource(s)",
            ResourceRelationType.OUTER_RESOURCE: "outer resource",
        }
        detail = (
            " ({})".format(self._relation_entity.metadata.get("id"))
            if verbose
            else ""
        )
        return "{type}{detail}".format(
            type=rel_type_map.get(self._relation_entity.type, "<unknown>"),
            detail=detail,
        )

    @property
    def detail(self) -> Sequence[str]:
        ent = self._relation_entity
        if ent.type is ResourceRelationType.ORDER:
            return _order_metadata_to_str(ent.metadata)
        if ent.type is ResourceRelationType.ORDER_SET:
            return _order_set_metadata_to_str(ent.metadata)
        if (
            ent.type is ResourceRelationType.INNER_RESOURCES
            and
            len(ent.members) > 1
        ):
            return ["members: {}".format(" ".join(ent.members))]
        return []


def _order_metadata_to_str(metadata: Mapping[str, Any]) -> Sequence[str]:
    return [
        "{action1} {resource1} then {action2} {resource2}".format(
            action1=metadata["first-action"],
            resource1=metadata["first"],
            action2=metadata["then-action"],
            resource2=metadata["then"],
        )
    ] + _order_common_metadata_to_str(metadata)


def _order_set_metadata_to_str(metadata: Mapping[str, Any]) -> Sequence[str]:
    result = []
    for res_set in metadata["sets"]:
        result.append("   set {resources}{options}".format(
            resources=" ".join(res_set["members"]),
            options=_resource_set_options_to_str(res_set["metadata"]),
        ))
    return _order_common_metadata_to_str(metadata) + result


def _resource_set_options_to_str(metadata: Mapping[str, Any]) -> str:
    supported_keys = (
        "sequential", "require-all", "ordering", "action", "role", "kind",
        "score",
    )
    result = _filter_supported_keys(metadata, supported_keys)
    return f" ({result})" if result else ""


def _filter_supported_keys(
    data: Mapping[str, Any], supported_keys: Iterable[str]
) -> str:
    return " ".join([
        f"{key}={value}"
        for key, value in sorted(data.items())
        if key in supported_keys
    ])


def _order_common_metadata_to_str(metadata: Mapping[str, Any]) -> List[str]:
    result = _filter_supported_keys(
        metadata, ("symmetrical", "kind", "require-all", "score")
    )
    return [result] if result else []



from pcs.common.fencing_topology import (
    TARGET_TYPE_REGEXP,
    TARGET_TYPE_ATTRIBUTE,
)
from pcs.common.tools import Version
from pcs.lib.cib import fencing_topology as cib_fencing_topology
from pcs.lib.cib.tools import (
    get_fencing_topology,
    get_resources,
)
from pcs.lib.pacemaker.live import get_cluster_status_xml
from pcs.lib.pacemaker.state import ClusterState

def add_level(
    lib_env, level, target_type, target_value, devices,
    force_device=False, force_node=False
):
    """
    Validate and add a new fencing level

    LibraryEnvironment lib_env -- environment
    int|string level -- level (index) of the new fencing level
    constant target_type -- the new fencing level target value type
    mixed target_value -- the new fencing level target value
    Iterable devices -- list of stonith devices for the new fencing level
    bool force_device -- continue even if a stonith device does not exist
    bool force_node -- continue even if a node (target) does not exist
    """
    version_check = None
    if target_type == TARGET_TYPE_REGEXP:
        version_check = Version(2, 3, 0)
    elif target_type == TARGET_TYPE_ATTRIBUTE:
        version_check = Version(2, 4, 0)

    cib = lib_env.get_cib(version_check)
    cib_fencing_topology.add_level(
        lib_env.report_processor,
        get_fencing_topology(cib),
        get_resources(cib),
        level,
        target_type,
        target_value,
        devices,
        ClusterState(
            get_cluster_status_xml(lib_env.cmd_runner())
        ).node_section.nodes,
        force_device,
        force_node
    )
    lib_env.report_processor.send()
    lib_env.push_cib()

def get_config(lib_env):
    """
    Get fencing levels configuration.

    Return a list of levels where each level is a dict with keys: target_type,
    target_value. level and devices. Devices is a list of stonith device ids.

    LibraryEnvironment lib_env -- environment
    """
    cib = lib_env.get_cib()
    return cib_fencing_topology.export(get_fencing_topology(cib))

def remove_all_levels(lib_env):
    """
    Remove all fencing levels
    LibraryEnvironment lib_env -- environment
    """
    cib_fencing_topology.remove_all_levels(
        get_fencing_topology(lib_env.get_cib())
    )
    lib_env.push_cib()

def remove_levels_by_params(
    lib_env, level=None, target_type=None, target_value=None, devices=None,
    ignore_if_missing=False
):
    """
    Remove specified fencing level(s)

    LibraryEnvironment lib_env -- environment
    int|string level -- level (index) of the fencing level to remove
    constant target_type -- the removed fencing level target value type
    mixed target_value -- the removed fencing level target value
    Iterable devices -- list of stonith devices of the removed fencing level
    bool ignore_if_missing -- when True, do not report if level not found
    """
    cib_fencing_topology.remove_levels_by_params(
        lib_env.report_processor,
        get_fencing_topology(lib_env.get_cib()),
        level,
        target_type,
        target_value,
        devices,
        ignore_if_missing
    )
    lib_env.report_processor.send()
    lib_env.push_cib()

def verify(lib_env):
    """
    Check if all cluster nodes and stonith devices used in fencing levels exist

    LibraryEnvironment lib_env -- environment
    """
    cib = lib_env.get_cib()
    cib_fencing_topology.verify(
        lib_env.report_processor,
        get_fencing_topology(cib),
        get_resources(cib),
        ClusterState(
            get_cluster_status_xml(lib_env.cmd_runner())
        ).node_section.nodes
    )
    lib_env.report_processor.send()



import base64
from collections import namedtuple

from pcs.lib import reports
from pcs.lib.errors import LibraryError

def create_pcmk_remote_actions(action_list):
    return {
        "pacemaker_remote {0}".format(action): service_cmd_format(
            "pacemaker_remote", action
        )
        for action in action_list
    }

def pcmk_authkey_format(authkey_content):
    """
    Return a dict usable in the communication with a remote/put_file
    authkey_content is raw authkey content
    """
    return {
        "data": base64.b64encode(authkey_content).decode("utf-8"),
        "type": "pcmk_remote_authkey",
        "rewrite_existing": True,
    }

def corosync_authkey_format(authkey_content):
    """
    Return a dict usable in the communication with a remote/put_file
    authkey_content is raw authkey content
    """
    return {
        "data": base64.b64encode(authkey_content).decode("utf-8"),
        "type": "corosync_authkey",
        "rewrite_existing": True,
    }

def pcmk_authkey_file(authkey_content):
    return {
        "pacemaker_remote authkey": pcmk_authkey_format(authkey_content)
    }

def corosync_authkey_file(authkey_content):
    return {
        "corosync authkey": corosync_authkey_format(authkey_content)
    }

def corosync_conf_format(corosync_conf_content):
    return {
        "type": "corosync_conf",
        "data": corosync_conf_content,
    }

def corosync_conf_file(corosync_conf_content):
    return {
        "corosync.conf": corosync_conf_format(corosync_conf_content)
    }

def pcs_settings_conf_format(content):
    return {
        "data": content,
        "type": "pcs_settings_conf",
        "rewrite_existing": True,
    }

def pcs_settings_conf_file(content):
    return {
        "pcs_settings.conf": pcs_settings_conf_format(content)
    }

def service_cmd_format(service, command):
    """
    Return a dict usable in the communication with a remote/run_action
    string service is name of requested service (eg. pacemaker_remote)
    string command specifies an action on service (eg. start)
    """
    return {
        "type": "service_command",
        "service": service,
        "command": command,
    }

class Result(namedtuple("Result", "code message")):
    """ Wrapper over some call results """

def unpack_items_from_response(main_response, main_key, node_label):
    """
    Check format of main_response and return main_response[main_key].
    dict main_response has on the key 'main_key' dict with item name as key and
        dict with result as value. E.g.
        {
            "files": {
                "file1": {"code": "success", "message": ""}
            }
        }
    string main_key is name of key under that is a dict with results
    string node_label is a node label for reporting an invalid format
    """
    is_in_expected_format = (
        isinstance(main_response, dict)
        and
        main_key in main_response
        and
        isinstance(main_response[main_key], dict)
    )

    if not is_in_expected_format:
        raise LibraryError(reports.invalid_response_format(node_label))

    return main_response[main_key]

def response_items_to_result(response_items, expected_keys, node_label):
    """
    Check format of response_items and return dict where keys are transformed to
    Result. E.g.
    {"file1": {"code": "success", "message": ""}}
    ->
    {"file1": Result("success", "")}}

    dict resposne_items has item name as key and dict with result as value.
    list expected_keys contains expected keys in a dict main_response[main_key]
    string node_label is a node label for reporting an invalid format
    """
    if set(expected_keys) != set(response_items.keys()):
        raise LibraryError(reports.invalid_response_format(node_label))

    for result in response_items.values():
        if(
            not isinstance(result, dict)
            or
            "code" not in result
            or
            "message" not in result
        ):
            raise LibraryError(reports.invalid_response_format(node_label))

    return {
        file_key: Result(raw_result["code"], raw_result["message"])
        for file_key, raw_result in response_items.items()
    }


def response_to_result(
    main_response, main_key, expected_keys, node_label
):
    """
    Validate response (from remote/put_file or remote/run_action) and transform
    results from dict to Result.

    dict main_response has on the key 'main_key' dict with item name as key and
        dict with result as value. E.g.
        {
            "files": {
                "file1": {"code": "success", "message": ""}
            }
        }
    string main_key is name of key under that is a dict with results
    list expected_keys contains expected keys in a dict main_response[main_key]
    string node_label is a node label for reporting an invalid format
    """
    return response_items_to_result(
        unpack_items_from_response(main_response, main_key, node_label),
        expected_keys,
        node_label
    )

def get_format_result(code_message_map):
    def format_result(result):
        if result.code in code_message_map:
            return code_message_map[result.code]

        return result.message
    return format_result
